"""Utilities for GCP instances."""
import copy
import enum
import functools
from multiprocessing import pool
import re
import subprocess
import time
from typing import Any, Dict, List, Optional, Tuple
import uuid

from sky import sky_logging
from sky.adaptors import gcp
from sky.clouds import gcp as gcp_cloud
from sky.provision import common
from sky.provision import constants as provision_constants
from sky.provision.gcp import constants
from sky.provision.gcp import mig_utils
from sky.utils import common_utils
from sky.utils import ux_utils

# Tag for the name of the node
INSTANCE_NAME_MAX_LEN = 64
INSTANCE_NAME_UUID_LEN = 8

TPU_NODE_CREATION_FAILURE = 'Failed to provision TPU node.'

# This is the maximum number of times we will retry a GCP API call.
# The number is identical to those we use for AWS boto3.
GCP_MAX_RETRIES = 12
GCP_CREATE_MAX_RETRIES = 5
GCP_RETRY_INTERVAL_SECONDS = 5
GCP_TIMEOUT = 300

logger = sky_logging.init_logger(__name__)

_FIREWALL_RESOURCE_NOT_FOUND_PATTERN = re.compile(
    r'The resource \'projects/.*/global/firewalls/.*\' was not found')


def _retry_on_gcp_http_exception(
    regex: Optional[str] = None,
    max_retries: int = GCP_MAX_RETRIES,
    retry_interval_s: int = GCP_RETRY_INTERVAL_SECONDS,
):
    """Retry a function call n-times for as long as it throws an exception."""

    def dec(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            def try_catch_exc():
                try:
                    value = func(*args, **kwargs)
                    return value
                except Exception as e:  # pylint: disable=broad-except
                    if (isinstance(e, gcp.http_error_exception()) and
                        (regex is None or re.search(regex, str(e)))):
                        logger.error(
                            f'Retrying for gcp.http_error_exception: {e}')
                        return e
                    raise

            for _ in range(max_retries):
                ret = try_catch_exc()
                if not isinstance(ret, Exception):
                    break
                logger.debug(f'Retrying for exception: {ret}')
                time.sleep(retry_interval_s)
            if isinstance(ret, Exception):
                raise ret
            return ret

        return wrapper

    return dec


def _generate_node_name(cluster_name: str, node_suffix: str,
                        is_head: bool) -> str:
    """Generate node name from labels and suffix.

    This is required so that the correct resource can be selected
    when the only information autoscaler has is the name of the node.

    The suffix is expected to be one of 'compute' or 'tpu'
    (as in ``GCPNodeType``).
    """
    suffix_id = common_utils.base36_encode(uuid.uuid4().hex)
    suffix = f'-{suffix_id[:INSTANCE_NAME_UUID_LEN]}-{node_suffix}'
    if is_head:
        suffix = f'-head{suffix}'
    else:
        suffix = f'-worker{suffix}'
    node_name = cluster_name + suffix
    assert len(node_name) <= INSTANCE_NAME_MAX_LEN, cluster_name
    return node_name


def _format_and_log_message_from_errors(errors: List[Dict[str, str]], e: Any,
                                        zone: Optional[str]) -> str:
    """Format errors into a string and log it to the console."""
    if errors:
        plural = 's' if len(errors) > 1 else ''
        codes = ', '.join(repr(e.get('code', 'N/A')) for e in errors)
        messages = '; '.join(
            repr(e.get('message', 'N/A').strip('.')) for e in errors)
        zone_str = f' in {zone}' if zone else ''
        msg = f'Got return code{plural} {codes}{zone_str}: {messages}'
    else:
        msg = f'create_instances: Failed with reason: {e}'
    logger.warning(msg)
    return msg


def selflink_to_name(selflink: str) -> str:
    """Converts a selflink to a name.

    Args:
        selflink: The selflink to convert.

    Returns:
        The name of the resource.
    """
    return selflink.rsplit('/', 1)[-1]


def instance_to_handler(instance: str):
    instance_type = instance.split('-')[-1]
    if instance_type == 'compute':
        return GCPComputeInstance
    elif instance_type == 'tpu':
        return GCPTPUVMInstance
    elif instance.startswith(constants.MIG_NAME_PREFIX):
        return GCPManagedInstanceGroup
    else:
        raise ValueError(f'Unknown instance type: {instance_type}')


class GCPInstance:
    """Base class for GCP instance handlers."""
    PENDING_STATES: List[str] = []
    NEED_TO_STOP_STATES: List[str] = []
    NON_STOPPED_STATES: List[str] = []
    NEED_TO_TERMINATE_STATES: List[str] = []
    RUNNING_STATE: str = ''
    STOPPING_STATES: List[str] = []
    STOPPED_STATES: List[str] = []
    STATUS_FIELD: str = ''

    @classmethod
    def load_resource(cls):
        """Load the GCP API for the instance type.

        Do not cache the resource object, as it will not work
        when multiple threads are running.
        """
        raise NotImplementedError

    @classmethod
    def stop(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> dict:
        raise NotImplementedError

    @classmethod
    def terminate(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> dict:
        raise NotImplementedError

    @classmethod
    def wait_for_operation(cls,
                           operation: dict,
                           project_id: str,
                           region: Optional[str] = None,
                           zone: Optional[str] = None) -> None:
        raise NotImplementedError

    @classmethod
    def filter(
        cls,
        project_id: str,
        zone: str,
        label_filters: Optional[Dict[str, str]],
        status_filters: Optional[List[str]],
        included_instances: Optional[List[str]] = None,
        excluded_instances: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def get_vpc_name(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> str:
        raise NotImplementedError

    @classmethod
    def delete_firewall_rule(
        cls,
        project_id: str,
        firewall_rule_name: str,
    ) -> None:
        raise NotImplementedError

    @classmethod
    def create_or_update_firewall_rule(
        cls,
        firewall_rule_name: str,
        project_id: str,
        vpc_name: str,
        cluster_name_on_cloud: str,
        ports: List[str],
    ) -> dict:
        raise NotImplementedError

    @classmethod
    def add_network_tag_if_not_exist(
        cls,
        project_id: str,
        zone: str,
        instance: str,
        tag: str,
    ) -> None:
        raise NotImplementedError

    @classmethod
    def create_instances(
        cls,
        cluster_name: str,
        project_id: str,
        zone: str,
        node_config: dict,
        labels: dict,
        count: int,
        total_count: int,
        include_head_node: bool,
    ) -> Tuple[Optional[List], List[str]]:
        """Creates multiple instances and returns result.

        Returns a tuple of (errors, list[instance_names]).
        """
        raise NotImplementedError

    @classmethod
    def start_instances(cls, cluster_name: str, project_id: str, zone: str,
                        instances: List[str], labels: Dict[str,
                                                           str]) -> List[str]:
        """Start multiple instances.

        Returns:
            List of instance names that are started.
        """
        del cluster_name  # Unused
        for instance_id in instances:
            cls.start_instance(instance_id, project_id, zone)
            cls.set_labels(project_id, zone, instance_id, labels)
        return instances

    @classmethod
    def start_instance(cls, node_id: str, project_id: str, zone: str) -> None:
        """Start a stopped instance."""
        raise NotImplementedError

    @classmethod
    def set_labels(cls, project_id: str, availability_zone: str, node_id: str,
                   labels: dict) -> None:
        raise NotImplementedError

    @classmethod
    def create_node_tag(cls,
                        project_id: str,
                        availability_zone: str,
                        target_instance_id: str,
                        is_head: bool = True) -> str:
        if is_head:
            node_tag = provision_constants.HEAD_NODE_TAGS
        else:
            node_tag = provision_constants.WORKER_NODE_TAGS
        cls.set_labels(project_id=project_id,
                       availability_zone=availability_zone,
                       node_id=target_instance_id,
                       labels=node_tag)

        return target_instance_id

    @classmethod
    def get_instance_info(cls, project_id: str, availability_zone: str,
                          instance_id: str) -> List[common.InstanceInfo]:
        raise NotImplementedError

    @classmethod
    def resize_disk(cls, project_id: str, availability_zone: str,
                    node_config: dict, instance_name: str) -> None:
        """Resize a Google Cloud disk based on the provided configuration.
        Returns the response of resize operation.
        """
        raise NotImplementedError


class GCPComputeInstance(GCPInstance):
    """Instance handler for GCP compute instances."""
    PENDING_STATES = ['PROVISIONING', 'STAGING', 'REPAIRING']
    STOPPING_STATES = ['STOPPING', 'SUSPENDING']
    STOPPED_STATES = ['TERMINATED', 'SUSPENDED']
    RUNNING_STATE = 'RUNNING'
    STATUS_FIELD = 'status'
    NEED_TO_STOP_STATES = PENDING_STATES + [RUNNING_STATE]

    NON_STOPPED_STATES = NEED_TO_STOP_STATES + STOPPING_STATES

    @classmethod
    def load_resource(cls):
        return gcp.build('compute',
                         'v1',
                         credentials=None,
                         cache_discovery=False)

    @classmethod
    def stop(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> dict:
        operation = cls.load_resource().instances().stop(
            project=project_id,
            zone=zone,
            instance=instance,
            # This is needed for the instance that has local SSDs attached by
            # default, such as a2-highgpu-8g. Otherwise, an error will be
            # raised. Refer to issue #2586
            # https://cloud.google.com/compute/docs/disks/local-ssd#stop_instance
            discardLocalSsd=False,
        ).execute()
        return operation

    @classmethod
    def terminate(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> dict:
        operation = cls.load_resource().instances().delete(
            project=project_id,
            zone=zone,
            instance=instance,
        ).execute()
        return operation

    @classmethod
    def filter(
        cls,
        project_id: str,
        zone: str,
        label_filters: Optional[Dict[str, str]],
        status_filters: Optional[List[str]],
        included_instances: Optional[List[str]] = None,
        excluded_instances: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        if label_filters:
            label_filter_expr = ('(' + ' AND '.join([
                '(labels.{key} = {value})'.format(key=key, value=value)
                for key, value in label_filters.items()
            ]) + ')')
        else:
            label_filter_expr = ''

        if status_filters:
            instance_state_filter_expr = ('(' + ' OR '.join([
                '(status = {status})'.format(status=status)
                for status in status_filters
            ]) + ')')
        else:
            instance_state_filter_expr = ''

        not_empty_filters = [
            f for f in [
                label_filter_expr,
                instance_state_filter_expr,
            ] if f
        ]

        filter_expr = ' AND '.join(not_empty_filters)

        response = (cls.load_resource().instances().list(
            project=project_id,
            filter=filter_expr,
            zone=zone,
        ).execute(num_retries=GCP_MAX_RETRIES))
        instances = response.get('items', [])
        instances = {i['name']: i for i in instances}
        if included_instances:
            instances = {
                k: v for k, v in instances.items() if k in included_instances
            }
        if excluded_instances:
            instances = {
                k: v
                for k, v in instances.items()
                if k not in excluded_instances
            }
        return instances

    @classmethod
    def wait_for_operation(cls,
                           operation: dict,
                           project_id: str,
                           region: Optional[str] = None,
                           zone: Optional[str] = None,
                           timeout: int = GCP_TIMEOUT) -> None:
        if zone is not None:
            kwargs = {'zone': zone}
            operation_caller = cls.load_resource().zoneOperations()
        elif region is not None:
            kwargs = {'region': region}
            operation_caller = cls.load_resource().regionOperations()
        else:
            kwargs = {}
            operation_caller = cls.load_resource().globalOperations()
        logger.debug(
            f'Waiting GCP operation {operation["name"]} to be ready ...')

        @_retry_on_gcp_http_exception(
            f'Failed to wait for operation {operation["name"]}')
        def call_operation(fn, timeout: int):
            request = fn(
                project=project_id,
                operation=operation['name'],
                **kwargs,
            )
            request.http.timeout = timeout
            return request.execute(num_retries=GCP_MAX_RETRIES)

        wait_start = time.time()
        while time.time() - wait_start < timeout:
            # Retry the wait() call until it succeeds or times out.
            # This is because the wait() call is only best effort, and does not
            # guarantee that the operation is done when it returns.
            # Reference: https://cloud.google.com/workflows/docs/reference/googleapis/compute/v1/zoneOperations/wait # pylint: disable=line-too-long
            remaining_timeout = max(timeout - (time.time() - wait_start), 1)
            result = call_operation(operation_caller.wait, remaining_timeout)
            if result['status'] == 'DONE':
                # NOTE: Error example:
                # {
                #   'code': 'VM_MIN_COUNT_NOT_REACHED',
                #   'message': 'Requested minimum count of 4 VMs could not be created.'
                # }
                errors = result.get('error', {}).get('errors')
                if errors is not None:
                    logger.debug(
                        'wait_operations: Failed to create instances. Reason: '
                        f'{errors}')
                    msg = _format_and_log_message_from_errors(
                        errors, result, zone)
                    error = common.ProvisionerError('Operation failed')
                    setattr(error, 'detailed_reason', msg)
                    error.errors = errors
                    raise error
                return
            logger.debug(f'wait_for_operation: Retry waiting for operation '
                         f'{operation["name"]} to finish (result: {result})...')
        else:
            logger.warning('wait_for_operation: Timeout waiting for creation '
                           'operation, cancelling the operation ...')
            remaining_timeout = max(timeout - (time.time() - wait_start), 1)
            try:
                result = call_operation(operation_caller.delete,
                                        remaining_timeout)
            except gcp.http_error_exception() as e:
                logger.debug('wait_for_operation: failed to cancel operation '
                             f'due to error: {e}')
            errors = [{
                'code': 'TIMEOUT',
                'message': f'Timeout waiting for operation {operation["name"]}',
                'domain': 'wait_for_operation'
            }]
            msg = _format_and_log_message_from_errors(errors, None, zone)
            error = common.ProvisionerError('Operation timed out')
            # Used for usage collection only, to include in the usage message.
            setattr(error, 'detailed_reason', msg)
            error.errors = errors
            raise error

    @classmethod
    def get_vpc_name(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> str:
        try:
            response = cls.load_resource().instances().get(
                project=project_id,
                zone=zone,
                instance=instance,
            ).execute()
            # Format: projects/PROJECT_ID/global/networks/VPC_NAME
            vpc_link = response['networkInterfaces'][0]['network']
            return selflink_to_name(vpc_link)
        except gcp.http_error_exception() as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to get VPC name for instance {instance}') from e

    @classmethod
    def add_network_tag_if_not_exist(
        cls,
        project_id: str,
        zone: str,
        instance: str,
        tag: str,
    ) -> None:
        try:
            # If we have multiple instances, they are in the same cluster,
            # i.e. the same VPC. So we can just pick one.
            response = cls.load_resource().instances().get(
                project=project_id,
                zone=zone,
                instance=instance,
            ).execute()
            existing_tags = response['tags'].get('items', [])
            if tag in existing_tags:
                return
            update_body = response['tags']
            update_body['items'] = existing_tags
            update_body['items'].append(tag)
            cls.load_resource().instances().setTags(
                project=project_id,
                zone=zone,
                instance=instance,
                body=update_body,
            ).execute()
        except gcp.http_error_exception() as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to add network tags for instance {instance}'
                ) from e

    @classmethod
    def delete_firewall_rule(
        cls,
        project_id: str,
        firewall_rule_name: str,
    ) -> None:
        rule = cls.load_resource().firewalls().list(
            project=project_id, filter=f'name={firewall_rule_name}').execute()
        # For the return value format, please refer to
        # https://developers.google.com/resources/api-libraries/documentation/compute/alpha/python/latest/compute_alpha.firewalls.html#list # pylint: disable=line-too-long
        if 'items' not in rule:
            logger.warning(f'Firewall rule {firewall_rule_name} not found. '
                           'Skip cleanup.')
            return
        cls.load_resource().firewalls().delete(
            project=project_id,
            firewall=firewall_rule_name,
        ).execute()

    @classmethod
    def create_or_update_firewall_rule(
        cls,
        firewall_rule_name: str,
        project_id: str,
        vpc_name: str,
        cluster_name_on_cloud: str,
        ports: List[str],
    ) -> dict:
        try:
            body = cls.load_resource().firewalls().get(
                project=project_id, firewall=firewall_rule_name).execute()
            body['allowed'][0]['ports'] = ports
            operation = cls.load_resource().firewalls().update(
                project=project_id,
                firewall=firewall_rule_name,
                body=body,
            ).execute()
        except gcp.http_error_exception() as e:
            if _FIREWALL_RESOURCE_NOT_FOUND_PATTERN.search(e.reason) is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Failed to update firewall rule {firewall_rule_name}'
                    ) from e
            body = {
                'name': firewall_rule_name,
                'description': (f'Allow user-specified port {ports} for '
                                f'cluster {cluster_name_on_cloud}'),
                'network': f'projects/{project_id}/global/networks/{vpc_name}',
                'selfLink': f'projects/{project_id}/global/firewalls/' +
                            firewall_rule_name,
                'direction': 'INGRESS',
                'priority': 65534,
                'allowed': [{
                    'IPProtocol': 'tcp',
                    'ports': ports,
                }],
                'sourceRanges': ['0.0.0.0/0'],
                'targetTags': [cluster_name_on_cloud],
            }
            operation = cls.load_resource().firewalls().insert(
                project=project_id,
                body=body,
            ).execute()
        return operation

    @classmethod
    # When there is a cloud function running in parallel to set labels for
    # newly created instances, it may fail with the following error:
    #   "Labels fingerprint either invalid or resource labels have changed"
    # We should retry until the labels are set successfully.
    @_retry_on_gcp_http_exception('Labels fingerprint either invalid')
    def set_labels(cls, project_id: str, availability_zone: str, node_id: str,
                   labels: dict) -> None:
        node = cls.load_resource().instances().get(
            project=project_id,
            instance=node_id,
            zone=availability_zone,
        ).execute(num_retries=GCP_CREATE_MAX_RETRIES)
        body = {
            'labels': dict(node['labels'], **labels),
            'labelFingerprint': node['labelFingerprint'],
        }
        operation = (cls.load_resource().instances().setLabels(
            project=project_id,
            zone=availability_zone,
            instance=node_id,
            body=body,
        ).execute(num_retries=GCP_CREATE_MAX_RETRIES))

        cls.wait_for_operation(operation, project_id, zone=availability_zone)

    @classmethod
    def create_instances(
        cls,
        cluster_name: str,
        project_id: str,
        zone: str,
        node_config: dict,
        labels: dict,
        count: int,
        total_count: int,
        include_head_node: bool,
    ) -> Tuple[Optional[List], List[str]]:
        # NOTE: The syntax for bulkInsert() is different from insert().
        # bulkInsert expects resource names without prefix. Otherwise
        # it causes a 503 error.
        config = copy.deepcopy(node_config)

        # removing TPU-specific default key set in config.py
        config.pop('networkConfig', None)

        head_tag_needed = [False] * count
        if include_head_node:
            head_tag_needed[0] = True

        names = []
        for i in range(count):
            names.append(
                _generate_node_name(cluster_name,
                                    GCPNodeType.COMPUTE.value,
                                    is_head=head_tag_needed[i]))

        labels = dict(config.get('labels', {}), **labels)

        config.update({
            'labels': dict(
                labels, **{
                    provision_constants.TAG_RAY_CLUSTER_NAME: cluster_name,
                    provision_constants.TAG_SKYPILOT_CLUSTER_NAME: cluster_name
                }),
        })

        all_names = []
        if 'reservationAffinity' in config:
            specific_reservations = set(config['reservationAffinity']['values'])
            reservations = gcp_cloud.GCP().get_reservations_available_resources(
                config['machineType'],
                region=zone.rpartition('-')[0],
                zone=zone,
                specific_reservations=specific_reservations)
            # Filter the reservations by the user-specified ones, because
            # reservations contain auto reservations as well, which do not
            # need to explicitly specify in the config for creating instances.
            specific_reservations_to_count = {
                reservation: count
                for reservation, count in reservations.items()
                if reservation in specific_reservations
            }
            # Sort the reservations by the number of available resources
            specified_reservations_list = sorted(
                specific_reservations_to_count.items(),
                key=lambda x: x[1],
                reverse=True)
            # TODO(zhwu): Convert this to parallel execution.
            # TODO(zhwu): This is not atomic as the reservation count may change
            # between the time we check and the time we create the instances, as
            # other users may be creating instances at the same time.
            # Our current implementation will skip the current region if the
            # reservation count is not enough, which is suboptimal.
            for reservation, reservation_count in specified_reservations_list:
                if reservation_count <= 0:
                    continue
                reservation_count = min(reservation_count, count)
                logger.debug(f'Creating {reservation_count} instances '
                             f'with reservation {reservation}')
                config['reservationAffinity']['values'] = [reservation]
                created_names = names[:reservation_count]
                errors = cls._create_instances(
                    created_names, project_id, zone, config,
                    head_tag_needed[:reservation_count])
                all_names.extend(created_names)
                if errors:
                    return errors, all_names
                count -= reservation_count
                if count <= 0:
                    return None, all_names
                names = names[reservation_count:]
                head_tag_needed = head_tag_needed[reservation_count:]
            config.pop('reservationAffinity', None)

        errors = cls._create_instances(names, project_id, zone, config,
                                       head_tag_needed)

        all_names.extend(names)
        return errors, all_names

    @classmethod
    def _insert(cls, names: List[str], project_id: str, zone: str,
                config: dict) -> List[dict]:
        # Convert name to selflink
        existing_machine_type = config['machineType']
        if not re.search('.*/machineTypes/.*', existing_machine_type):
            config['machineType'] = (
                f'zones/{zone}/machineTypes/{config["machineType"]}')

        for accelerator in config.get('guestAccelerators', []):
            gpu_type = accelerator['acceleratorType']
            if not re.search('.*/acceleratorTypes/.*', gpu_type):
                accelerator['acceleratorType'] = (
                    f'projects/{project_id}/zones/{zone}/'
                    f'acceleratorTypes/{gpu_type}')

        logger.debug('Launching GCP instances with "insert" ...')
        operations = []
        for name in names:
            body = {
                'name': name,
                **config,
            }
            request = cls.load_resource().instances().insert(
                project=project_id,
                zone=zone,
                body=body,
            )
            # We need to retry the insert operation because it may fail with
            # RESOURCE_OPERATION_RATE_EXCEEDED error, which is normally caused
            # by creating VMs with machine images on different zones.
            operation = request.execute(num_retries=GCP_MAX_RETRIES)
            operations.append(operation)

        logger.debug('"insert" operation requested ...')
        return operations

    @classmethod
    def _convert_selflinks_in_config(cls, config: dict) -> None:
        """Convert selflinks to names in the config."""
        for disk in config.get('disks', []):
            disk_type = disk.get('initializeParams', {}).get('diskType')
            if disk_type is not None:
                disk['initializeParams']['diskType'] = selflink_to_name(
                    disk_type)
        config['machineType'] = selflink_to_name(config['machineType'])
        for accelerator in config.get('guestAccelerators', []):
            accelerator['acceleratorType'] = selflink_to_name(
                accelerator['acceleratorType'])

    @classmethod
    def _bulk_insert(cls, names: List[str], project_id: str, zone: str,
                     config: dict) -> List[dict]:
        source_instance_template = config.pop('sourceInstanceTemplate', None)
        if 'scheduling' in config and isinstance(config['scheduling'], list):
            # For backeward compatibility: converting the list of dictionaries
            # to a dictionary due to the use of deprecated API.
            # [{'preemptible': True}, {'onHostMaintenance': 'TERMINATE'}]
            # to {'preemptible': True, 'onHostMaintenance': 'TERMINATE'}
            config['scheduling'] = {
                k: v for d in config['scheduling'] for k, v in d.items()
            }

        cls._convert_selflinks_in_config(config)

        body = {
            'count': len(names),
            'instanceProperties': config,
            'sourceInstanceTemplate': source_instance_template,
            'perInstanceProperties': {n: {} for n in names}
        }
        logger.debug('Launching GCP instances with "bulkInsert" ...')
        request = cls.load_resource().instances().bulkInsert(
            project=project_id,
            zone=zone,
            body=body,
        )
        operation = request.execute(num_retries=0)
        return [operation]

    @classmethod
    def _use_bulk_insert(cls, config) -> bool:
        """Decide whether to use bulkInsert or not based on config."""
        # bulkInsert does not support overriding parameter sourceMachineImage
        # with disks (without a sourceImage), causing the following error:
        #   'Invalid value for field \'resource.instanceProperties.disks[0]\':
        #   \'{  "type": "PERSISTENT",  "boot": true,  "initializeParams": {
        #    "diskSizeGb": "256",    "diskType"...\'. Boot disk must have a
        #    source specified'
        # https://cloud.google.com/compute/docs/reference/rest/v1/instances/bulkInsert # pylint: disable=line-too-long
        if config.get('sourceMachineImage') is not None:
            return False
        # bulkInsert does not support attaching existing
        # disks to the instances with READ_WRITE mode.
        if config.get('disks') is not None:
            for disk in config['disks']:
                if disk.get('source') is not None and disk.get(
                        'mode', 'READ_WRITE') == 'READ_WRITE':
                    return False
                if disk.get('initializeParams') is not None and disk.get(
                        'initializeParams', {}).get('diskName') is not None:
                    return False
        return True

    @classmethod
    def _create_instances(
        cls,
        names: List[str],
        project_id: str,
        zone: str,
        config: dict,
        head_tag_needed: List[bool],
    ) -> Optional[List]:

        def _handle_http_error(e):
            # NOTE: Error example:
            # {
            #   'message': "Quota '...' exceeded. Limit: ... in region xx-xxxx.", # pylint: disable=line-too-long
            #   'domain': 'usageLimits',
            #   'reason': 'quotaExceeded'
            # }
            error_details = getattr(e, 'error_details', [])
            errors = []
            for detail in error_details:
                # To be consistent with error messages returned by operation wait.
                errors.append({
                    'code': detail.get('reason'),
                    'domain': detail.get('domain'),
                    'message': detail.get('message', str(e)),
                })
            logger.debug(
                f'create_instances: googleapiclient.errors.HttpError: {e}')
            _format_and_log_message_from_errors(errors, e, zone)
            return errors

        # Allow Google Compute Engine instance templates.
        #
        # Config example:
        #
        #     ...
        #     node_config:
        #         sourceInstanceTemplate: global/instanceTemplates/worker-16
        #         machineType: e2-standard-16
        #     ...
        #
        # node_config parameters override matching template parameters, if any.
        #
        # https://cloud.google.com/compute/docs/instance-templates
        # https://cloud.google.com/compute/docs/reference/rest/v1/instances/insert
        try:
            if cls._use_bulk_insert(config):
                operations = cls._bulk_insert(names, project_id, zone, config)
            else:
                operations = cls._insert(names, project_id, zone, config)
        except gcp.http_error_exception() as e:
            return _handle_http_error(e)

        for operation in operations:
            errors = operation.get('error', {}).get('errors', [])
            if errors:
                logger.debug('create_instances: Failed to create instances. '
                             f'Reason: {errors}')
                _format_and_log_message_from_errors(errors, operations, zone)
                return errors

        logger.debug('Waiting GCP instances to be ready ...')
        try:
            for operation in operations:
                cls.wait_for_operation(operation, project_id, zone=zone)
        except common.ProvisionerError as e:
            return e.errors
        except gcp.http_error_exception() as e:
            return _handle_http_error(e)

        # assign labels for head node
        with pool.ThreadPool() as p:
            p.starmap(cls.create_node_tag,
                      [(project_id, zone, names[i], head_tag_needed[i])
                       for i in range(len(names))])
        return None

    @classmethod
    def start_instance(cls, node_id: str, project_id: str, zone: str) -> None:
        operation = (cls.load_resource().instances().start(
            project=project_id,
            zone=zone,
            instance=node_id,
        ).execute())

        cls.wait_for_operation(operation, project_id, zone=zone)

    @classmethod
    def get_instance_info(cls, project_id: str, availability_zone: str,
                          instance_id: str) -> List[common.InstanceInfo]:
        result = cls.load_resource().instances().get(
            project=project_id,
            zone=availability_zone,
            instance=instance_id,
        ).execute()
        external_ip = (result.get('networkInterfaces',
                                  [{}])[0].get('accessConfigs',
                                               [{}])[0].get('natIP', None))
        internal_ip = result.get('networkInterfaces', [{}])[0].get('networkIP')

        return [
            common.InstanceInfo(
                instance_id=instance_id,
                internal_ip=internal_ip,
                external_ip=external_ip,
                tags=result.get('labels', {}),
            )
        ]

    @classmethod
    def resize_disk(cls, project_id: str, availability_zone: str,
                    node_config: dict, instance_name: str) -> None:
        """Resize a Google Cloud disk based on the provided configuration."""

        # Extract the specified disk size from the configuration
        new_size_gb = node_config['disks'][0]['initializeParams']['diskSizeGb']

        # Fetch the instance details to get the disk name and current disk size
        response = (cls.load_resource().instances().get(
            project=project_id,
            zone=availability_zone,
            instance=instance_name,
        ).execute())
        disk_name = selflink_to_name(response['disks'][0]['source'])

        try:
            # Execute the resize request and return the response
            operation = (cls.load_resource().disks().resize(
                project=project_id,
                zone=availability_zone,
                disk=disk_name,
                body={
                    'sizeGb': str(new_size_gb),
                },
            ).execute())
        except gcp.http_error_exception() as e:
            # Catch HttpError when provided with invalid value for new disk
            # size. Allowing users to create instances with the same size as the
            # image.
            # TODO(zhwu): We should only match the error message that are using
            # the disk with same size as the image.
            logger.warning(f'googleapiclient.errors.HttpError: {e.reason}')
            return

        cls.wait_for_operation(operation, project_id, zone=availability_zone)


class GCPManagedInstanceGroup(GCPComputeInstance):
    """Handler for GCP Managed Instance Group."""

    @classmethod
    def create_instances(
        cls,
        cluster_name: str,
        project_id: str,
        zone: str,
        node_config: dict,
        labels: dict,
        count: int,
        total_count: int,
        include_head_node: bool,
    ) -> Tuple[Optional[List], List[str]]:
        logger.debug(f'Creating cluster with MIG: {cluster_name!r}')
        config = copy.deepcopy(node_config)
        labels = dict(config.get('labels', {}), **labels)

        config.update({
            'labels': dict(
                labels,
                **{
                    provision_constants.TAG_RAY_CLUSTER_NAME: cluster_name,
                    # Assume all nodes are workers, we can update the head node
                    # once the instances are created.
                    **provision_constants.WORKER_NODE_TAGS,
                    provision_constants.TAG_SKYPILOT_CLUSTER_NAME: cluster_name,
                }),
        })
        cls._convert_selflinks_in_config(config)

        # Convert label values to string and lowercase per MIG API requirement.
        region = zone.rpartition('-')[0]
        instance_template_name = mig_utils.get_instance_template_name(
            cluster_name)
        managed_instance_group_name = mig_utils.get_managed_instance_group_name(
            cluster_name)

        instance_template_exists = mig_utils.check_instance_template_exits(
            project_id, region, instance_template_name)
        mig_exists = mig_utils.check_managed_instance_group_exists(
            project_id, zone, managed_instance_group_name)

        label_filters = {
            provision_constants.TAG_RAY_CLUSTER_NAME: cluster_name,
        }
        potential_head_instances = []
        if mig_exists:
            instances = cls.filter(
                project_id,
                zone,
                label_filters={
                    provision_constants.TAG_RAY_NODE_KIND: 'head',
                    **label_filters,
                },
                status_filters=cls.NEED_TO_TERMINATE_STATES)
            potential_head_instances = list(instances.keys())

        config['labels'] = {
            k: str(v).lower() for k, v in config['labels'].items()
        }
        if instance_template_exists:
            if mig_exists:
                logger.debug(
                    f'Instance template {instance_template_name} already '
                    'exists. Skip creating it.')
            else:
                logger.debug(
                    f'Instance template {instance_template_name!r} '
                    'exists and no instance group is using it. This is a '
                    'leftover of a previous autodown. Delete it and recreate '
                    'it.')
                # TODO(zhwu): this is a bit hacky as we cannot delete instance
                # template during an autodown, we can only defer the deletion
                # to the next launch of a cluster with the same name. We should
                # find a better way to handle this.
                cls._delete_instance_template(project_id, zone,
                                              instance_template_name)
                instance_template_exists = False

        if not instance_template_exists:
            operation = mig_utils.create_region_instance_template(
                cluster_name, project_id, region, instance_template_name,
                config)
            cls.wait_for_operation(operation, project_id, region=region)
        # create managed instance group
        instance_template_url = (f'projects/{project_id}/regions/{region}/'
                                 f'instanceTemplates/{instance_template_name}')
        if not mig_exists:
            # Create a new MIG with size 0 and resize it later for triggering
            # DWS, according to the doc: https://cloud.google.com/compute/docs/instance-groups/create-mig-with-gpu-vms # pylint: disable=line-too-long
            operation = mig_utils.create_managed_instance_group(
                project_id,
                zone,
                managed_instance_group_name,
                instance_template_url,
                size=0)
            cls.wait_for_operation(operation, project_id, zone=zone)

        managed_instance_group_config = config[
            constants.MANAGED_INSTANCE_GROUP_CONFIG]
        if count > 0:
            # Use resize to trigger DWS for creating VMs.
            operation = mig_utils.resize_managed_instance_group(
                project_id,
                zone,
                managed_instance_group_name,
                count,
                run_duration=managed_instance_group_config['run_duration'])
            cls.wait_for_operation(operation, project_id, zone=zone)

        # This will block the provisioning until the nodes are ready, which
        # makes the failover not effective. We rely on the request timeout set
        # by user to trigger failover.
        mig_utils.wait_for_managed_group_to_be_stable(
            project_id,
            zone,
            managed_instance_group_name,
            timeout=managed_instance_group_config.get(
                'provision_timeout',
                constants.DEFAULT_MANAGED_INSTANCE_GROUP_PROVISION_TIMEOUT))

        pending_running_instance_names = cls._add_labels_and_find_head(
            cluster_name, project_id, zone, labels, potential_head_instances)
        assert len(pending_running_instance_names) == total_count, (
            pending_running_instance_names, total_count)
        cls.create_node_tag(
            project_id,
            zone,
            pending_running_instance_names[0],
            is_head=True,
        )
        return None, pending_running_instance_names

    @classmethod
    def _delete_instance_template(cls, project_id: str, zone: str,
                                  instance_template_name: str) -> None:
        logger.debug(f'Deleting instance template {instance_template_name}...')
        region = zone.rpartition('-')[0]
        try:
            operation = cls.load_resource().regionInstanceTemplates().delete(
                project=project_id,
                region=region,
                instanceTemplate=instance_template_name).execute()
            cls.wait_for_operation(operation, project_id, region=region)
        except gcp.http_error_exception() as e:
            if re.search(mig_utils.IT_RESOURCE_NOT_FOUND_PATTERN,
                         str(e)) is None:
                raise
            logger.debug(
                f'Instance template {instance_template_name!r} does not exist. '
                'Skip deletion.')

    @classmethod
    def delete_mig(cls, project_id: str, zone: str, cluster_name: str) -> bool:
        """Returns whether the MIG is deleted successfully."""
        mig_exists_and_deleted = True
        mig_name = mig_utils.get_managed_instance_group_name(cluster_name)
        # Get all resize request of the MIG and cancel them.
        mig_utils.cancel_all_resize_request_for_mig(project_id, zone, mig_name)
        logger.debug(f'Deleting MIG {mig_name!r} ...')
        try:
            operation = cls.load_resource().instanceGroupManagers().delete(
                project=project_id, zone=zone,
                instanceGroupManager=mig_name).execute()
            cls.wait_for_operation(operation, project_id, zone=zone)
        except gcp.http_error_exception() as e:
            if re.search(mig_utils.MIG_RESOURCE_NOT_FOUND_PATTERN,
                         str(e)) is None:
                raise
            logger.debug(f'MIG {mig_name!r} does not exist. Skip '
                         'deletion.')
            mig_exists_and_deleted = False

        # In the autostop case, the following deletion of instance template
        # will not be executed as the instance that runs the deletion will be
        # terminated with the managed instance group. It is ok to leave the
        # instance template there as when a user creates a new cluster with the
        # same name, the instance template will be updated in our
        # create_instances method.
        cls._delete_instance_template(
            project_id, zone,
            mig_utils.get_instance_template_name(cluster_name))
        return mig_exists_and_deleted

    @classmethod
    def _add_labels_and_find_head(
            cls, cluster_name: str, project_id: str, zone: str,
            labels: Dict[str, str],
            potential_head_instances: List[str]) -> List[str]:
        pending_running_instances = cls.filter(
            project_id,
            zone,
            {provision_constants.TAG_RAY_CLUSTER_NAME: cluster_name},
            # Find all provisioning and running instances.
            status_filters=cls.NEED_TO_STOP_STATES)
        for running_instance_name in pending_running_instances.keys():
            if running_instance_name in potential_head_instances:
                head_instance_name = running_instance_name
                break
        else:
            head_instance_name = list(pending_running_instances.keys())[0]
        # We need to update the node's label if mig already exists, as the
        # config is not updated during the resize operation.
        for instance_name in pending_running_instances.keys():
            cls.set_labels(project_id=project_id,
                           availability_zone=zone,
                           node_id=instance_name,
                           labels=labels)

        pending_running_instance_names = list(pending_running_instances.keys())
        pending_running_instance_names.remove(head_instance_name)
        # Label for head node type will be set by caller
        return [head_instance_name] + pending_running_instance_names


class GCPTPUVMInstance(GCPInstance):
    """Instance handler for GCP TPU VM."""
    PENDING_STATES = ['CREATING', 'STARTING', 'RESTARTING', 'REPAIRING']
    RUNNING_STATE = 'READY'
    STOPPING_STATES = ['STOPPING']
    STOPPED_STATES = ['STOPPED']
    STATUS_FIELD = 'state'
    NEED_TO_STOP_STATES = PENDING_STATES + [RUNNING_STATE]

    NON_STOPPED_STATES = NEED_TO_STOP_STATES + STOPPING_STATES

    @classmethod
    def load_resource(cls):
        return gcp.build(
            'tpu',
            constants.TPU_VM_VERSION,
            credentials=None,
            cache_discovery=False,
            discoveryServiceUrl='https://tpu.googleapis.com/$discovery/rest')

    @classmethod
    def wait_for_operation(cls,
                           operation: dict,
                           project_id: str,
                           region: Optional[str] = None,
                           zone: Optional[str] = None) -> None:
        """Poll for TPU operation until finished."""
        del project_id, region, zone  # unused

        @_retry_on_gcp_http_exception(
            f'Failed to wait for operation {operation["name"]}')
        def call_operation(fn, timeout: int):
            request = fn(name=operation['name'])
            request.http.timeout = timeout
            return request.execute(num_retries=GCP_MAX_RETRIES)

        wait_start = time.time()
        while time.time() - wait_start < GCP_TIMEOUT:
            timeout = max(GCP_TIMEOUT - (time.time() - wait_start), 1)
            result = call_operation(
                cls.load_resource().projects().locations().operations().get,
                timeout)
            if result['done']:
                break
            logger.debug('wait_for_tpu_operation: '
                         f'Waiting for operation {operation["name"]} to '
                         'finish ...')

        if 'error' in result:
            error = common.ProvisionerError('Operation failed')
            errors = []
            errors.append({
                'code': result['error']['code'],
                'message': result['error']['message'],
                'domain': 'wait_for_operation',
            })
            for detail in result['error'].get('details', []):
                errors.append({
                    'code': detail.pop('@type', ''),
                    'domain': 'wait_for_operation',
                    'message': str(detail),
                })
            error.errors = errors
            raise error

        if 'response' in result:
            logger.debug('wait_for_tpu_operation: '
                         f'Operation {operation["name"]} finished.')

    @classmethod
    def filter(
        cls,
        project_id: str,
        zone: str,
        label_filters: Optional[Dict[str, str]],
        status_filters: Optional[List[str]],
        included_instances: Optional[List[str]] = None,
        excluded_instances: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        path = f'projects/{project_id}/locations/{zone}'
        try:
            response = (cls.load_resource().projects().locations().nodes().list(
                parent=path).execute(num_retries=GCP_MAX_RETRIES))
        except gcp.http_error_exception() as e:
            # SKY: Catch HttpError when accessing unauthorized region.
            # Return empty dict instead of raising exception to not break.
            if 'is not found or access is unauthorized.' in str(e):
                return {}
            if 'Permission \'tpu.nodes.list\' denied on' in str(e):
                return {}
            logger.debug(f'filter: googleapiclient.errors.HttpError: {e}')
            raise

        instances = response.get('nodes', [])

        # filter_expr cannot be passed directly to API
        # so we need to filter the results ourselves

        # same logic as in GCPCompute.list_instances
        label_filters = label_filters or {}

        def filter_instance(instance) -> bool:
            labels = instance.get('labels', {})
            if label_filters:
                for key, value in label_filters.items():
                    if key not in labels:
                        return False
                    if value != labels[key]:
                        return False

            if status_filters:
                if instance.get('state') not in status_filters:
                    return False

            return True

        instances = list(filter(filter_instance, instances))
        instances = {i['name']: i for i in instances}

        if included_instances:
            instances = {
                k: v for k, v in instances.items() if k in included_instances
            }
        if excluded_instances:
            instances = {
                k: v
                for k, v in instances.items()
                if k not in excluded_instances
            }
        return instances

    @classmethod
    def stop(cls, project_id: str, zone: str, instance: str) -> dict:
        """Stop a TPU VM."""
        del project_id, zone  # unused
        operation = cls.load_resource().projects().locations().nodes().stop(
            name=instance).execute()
        return operation

    @classmethod
    def terminate(cls, project_id: str, zone: str, instance: str) -> dict:
        """Terminate a TPU VM."""
        del project_id, zone  # unused
        operation = cls.load_resource().projects().locations().nodes().delete(
            name=instance).execute()
        return operation

    @classmethod
    def add_network_tag_if_not_exist(
        cls,
        project_id: str,
        zone: str,
        instance: str,
        tag: str,
    ) -> None:
        # https://cloud.google.com/tpu/docs/reference/rest/v2alpha1/projects.locations.nodes  # pylint: disable=line-too-long
        # https://cloud.google.com/tpu/docs/reference/rest/v2alpha1/projects.locations.nodes/patch  # pylint: disable=line-too-long
        del project_id, zone  # unused
        try:
            response = cls.load_resource().projects().locations().nodes().get(
                name=instance).execute()
            existing_tags = response.get('tags', [])
            if tag in existing_tags:
                return
            existing_tags.append(tag)
            update_body = response
            update_body['tags'] = existing_tags
            cls.load_resource().projects().locations().nodes().patch(
                name=instance,
                body=update_body,
                updateMask='tags',
            ).execute()
        except gcp.http_error_exception() as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to add network tags for instance {instance}'
                ) from e

    @classmethod
    def get_vpc_name(
        cls,
        project_id: str,
        zone: str,
        instance: str,
    ) -> str:
        del project_id, zone  # unused
        try:
            response = cls.load_resource().projects().locations().nodes().get(
                name=instance).execute()
            vpc_link = response['networkConfig']['network']
            return selflink_to_name(vpc_link)
        except gcp.http_error_exception() as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Failed to get VPC name for instance {instance}') from e

    @classmethod
    @_retry_on_gcp_http_exception('unable to queue the operation')
    def set_labels(cls, project_id: str, availability_zone: str, node_id: str,
                   labels: dict) -> None:
        while True:
            # wait until the instance become ready before setting labels
            # as Cloud TPU API does not allow setting labels on pending
            # instances
            instances = cls.filter(
                project_id=project_id,
                zone=availability_zone,
                label_filters=None,
                status_filters=cls.PENDING_STATES,
                included_instances=[node_id],
            )
            if not instances:
                break
            logger.debug(f'set_labels: Waiting for instance {node_id} to be '
                         'ready...')
            time.sleep(constants.POLL_INTERVAL)

        node = (cls.load_resource().projects().locations().nodes().get(
            name=node_id).execute(num_retries=GCP_CREATE_MAX_RETRIES))
        body = {
            'labels': dict(node['labels'], **labels),
        }
        update_mask = 'labels'

        operation = (cls.load_resource().projects().locations().nodes().patch(
            name=node_id,
            updateMask=update_mask,
            body=body,
        ).execute(num_retries=GCP_CREATE_MAX_RETRIES))

        cls.wait_for_operation(operation, project_id, availability_zone)

    @classmethod
    def create_instances(
        cls,
        cluster_name: str,
        project_id: str,
        zone: str,
        node_config: dict,
        labels: dict,
        count: int,
        total_count: int,
        include_head_node: bool,
    ) -> Tuple[Optional[List], List[str]]:
        config = copy.deepcopy(node_config)
        # removing Compute-specific default key set in config.py
        config.pop('networkInterfaces', None)

        head_tag_needed = [False] * count
        if include_head_node:
            head_tag_needed[0] = True

        names = []
        for i in range(count):
            names.append(
                _generate_node_name(cluster_name,
                                    GCPNodeType.TPU.value,
                                    is_head=head_tag_needed[i]))

        labels = dict(config.get('labels', {}), **labels)

        config.update({
            'labels': dict(
                labels, **{
                    provision_constants.TAG_RAY_CLUSTER_NAME: cluster_name,
                    provision_constants.TAG_SKYPILOT_CLUSTER_NAME: cluster_name
                }),
        })

        if 'reservationAffinity' in config:
            raise NotImplementedError(
                'TPU VMs do not support reservations yet.')

        # Allow Google Compute Engine instance templates.
        #
        # Config example:
        #
        #     ...
        #     node_config:
        #         sourceInstanceTemplate: global/instanceTemplates/worker-16
        #         machineType: e2-standard-16
        #     ...
        #
        # node_config parameters override matching template parameters, if any.
        #
        # https://cloud.google.com/compute/docs/instance-templates
        # https://cloud.google.com/compute/docs/reference/rest/v1/instances/insert
        operations = []
        for i, name in enumerate(names):
            node_config = config.copy()
            if i == 0:
                node_config['labels'].update(provision_constants.HEAD_NODE_TAGS)
            else:
                node_config['labels'].update(
                    provision_constants.WORKER_NODE_TAGS)
            try:
                logger.debug('Launching GCP TPU VM ...')
                request = (
                    cls.load_resource().projects().locations().nodes().create(
                        parent=f'projects/{project_id}/locations/{zone}',
                        body=node_config,
                        nodeId=name,
                    ))
                operation = request.execute(num_retries=0)
                operations.append(operation)
            except gcp.http_error_exception() as e:
                # NOTE: Error example:
                # {
                #   'message': "Quota '...' exceeded. Limit: ... in region xx-xxxx.", # pylint: disable=line-too-long
                #   'domain': 'usageLimits',
                #   'reason': 'quotaExceeded'
                # }
                error_details = getattr(e, 'error_details', [])
                logger.debug(
                    f'create_instances: googleapiclient.errors.HttpError: {e}')
                errors = []
                if isinstance(error_details, str):
                    errors.append({
                        'code': 'CREATION_FAILED',
                        'domain': 'create_instances',
                        'message': error_details,
                    })
                    _format_and_log_message_from_errors(errors, e, zone)
                    return errors, names
                for detail in error_details:
                    # To be consistent with error messages returned by operation
                    # wait.
                    viloations = detail.get('violations', [])
                    if not viloations:
                        errors.append({
                            'code': detail.get('reason'),
                            'domain': detail.get('domain'),
                            'message': detail.get('message', str(e)),
                        })
                    else:
                        for violation in viloations:
                            errors.append({
                                'code': detail.get('@type'),
                                'domain': violation.get('subject'),
                                'message': violation.get('description'),
                            })
                _format_and_log_message_from_errors(errors, e, zone)
                return errors, names
        errors = []
        for operation in operations:
            error = operation.get('error')
            if error:
                error['domain'] = 'create_instances'
                errors.append(error)
            details = operation.get('error', {}).get('details', [])
            for detail in details:
                detail['code'] = detail.pop('@type', '')
                detail['message'] = str(detail)
                detail['domain'] = 'create_instances'
            if details:
                errors.extend(details)
        if errors:
            logger.debug('create_instances: Failed to create instances. '
                         f'Reason: {errors}')
            _format_and_log_message_from_errors(errors, operations, zone)
            return errors, names

        logger.debug('Waiting GCP instances to be ready ...')
        wait_start = time.time()
        success = [False] * len(operations)
        results: List[dict] = [{} for _ in range(len(operations))]
        while time.time() - wait_start < GCP_TIMEOUT:
            # Retry the wait() call until it succeeds or times out.
            # This is because the wait() call is only best effort, and does not
            # guarantee that the operation is done when it returns.
            # Reference: https://cloud.google.com/workflows/docs/reference/googleapis/compute/v1/zoneOperations/wait # pylint: disable=line-too-long
            for i, operation in enumerate(operations):
                if success[i]:
                    continue
                request = (
                    cls.load_resource().projects().locations().operations().get(
                        name=operation['name'],))
                request.http.timeout = GCP_TIMEOUT - (time.time() - wait_start)
                result = request.execute(num_retries=GCP_CREATE_MAX_RETRIES)
                results[i] = result
                success[i] = result['done']
            if all(success):
                logger.debug(f'create_instances: Finished {results}')
                break
            logger.debug('create_instances: Retry waiting for TPU operations '
                         f'to finish: {results}...')
        else:
            logger.warning('create_instances: Timeout waiting for TPU creation '
                           'operation, cancelling the operation ...')
            for i, operation in enumerate(operations):
                if success[i]:
                    continue
                request = cls.load_resource().projects().locations().operations(
                ).cancel(name=operation['name'],)
            request.http.timeout = GCP_TIMEOUT - (time.time() - wait_start)
            request.execute(num_retries=GCP_CREATE_MAX_RETRIES)
            errors = [{
                'code': 'TIMEOUT',
                'message': 'Timeout waiting for creation operation',
                'domain': 'create_instances'
            }]
            _format_and_log_message_from_errors(errors, None, zone)
            return errors, names

        # NOTE: Error example:
        # {
        #    'code': 8,
        #    'message': 'There is no more capacity in the zone ...
        # }
        errors = []
        for result in results:
            error = result.get('error', {})
            if error:
                errors.append(error)
        if errors:
            logger.debug(
                'create_instances: Failed to create instances. Reason: '
                f'{errors}')
            _format_and_log_message_from_errors(errors, results, zone)
            return errors, names
        assert all(success), (
            'Failed to create instances, but there is no error. '
            f'Instance status: {results}')
        return None, names

    @classmethod
    def start_instance(cls, node_id: str, project_id: str, zone: str) -> None:
        operation = (cls.load_resource().projects().locations().nodes().start(
            name=node_id).execute())

        cls.wait_for_operation(operation, project_id, zone)

    @classmethod
    def resize_disk(cls, project_id: str, availability_zone: str,
                    node_config: dict, instance_name: str) -> None:
        """Resize the disk a machine image with a different size is used.

        TODO: Implement the feature to attach persistent disks for TPU VMs.
        The boot disk of TPU VMs is not resizable, and users need to add a
        persistent disk to expand disk capacity. Related issue: #2387
        """
        return

    @classmethod
    def get_instance_info(cls, project_id: str, availability_zone: str,
                          instance_id: str) -> List[common.InstanceInfo]:
        del project_id, availability_zone  # unused
        result = cls.load_resource().projects().locations().nodes().get(
            name=instance_id).execute()
        network_endpoints = result.get('networkEndpoints', [{}])
        external_ips = []
        internal_ips = []
        for endpoint in network_endpoints:
            external_ips.append(
                endpoint.get('accessConfig', {}).get('externalIp', None))
            internal_ips.append(endpoint.get('ipAddress', None))

        return [
            common.InstanceInfo(
                instance_id=instance_id,
                internal_ip=internal_ip,
                external_ip=external_ip,
                tags=result.get('labels', {}),
            ) for internal_ip, external_ip in zip(internal_ips, external_ips)
        ]


class GCPNodeType(enum.Enum):
    """Enum for GCP node types (compute & tpu)"""

    COMPUTE = 'compute'
    MIG = 'mig'
    TPU = 'tpu'


def get_node_type(config: Dict[str, Any]) -> GCPNodeType:
    """Returns node type based on the keys in ``node``.

    This is a very simple check. If we have a ``machineType`` key,
    this is a Compute instance. If we don't have a ``machineType`` key,
    but we have ``acceleratorType``, this is a TPU. Otherwise, it's
    invalid and an exception is raised.

    This works for both node configs and API returned nodes.
    """
    if ('machineType' not in config and 'acceleratorType' not in config):
        raise ValueError(
            'Invalid node. For a Compute instance, "machineType" is '
            'required. '
            'For a TPU instance, "acceleratorType" and no "machineType" '
            'is required. '
            f'Got {list(config)}')

    if 'machineType' not in config and 'acceleratorType' in config:
        return GCPNodeType.TPU

    if (config.get(constants.MANAGED_INSTANCE_GROUP_CONFIG, None) is not None
            and config.get('guestAccelerators', None) is not None):
        # DWS in MIG only works for machine with GPUs.
        return GCPNodeType.MIG

    return GCPNodeType.COMPUTE


def create_tpu_node(project_id: str, zone: str, tpu_node_config: Dict[str, str],
                    vpc_name: str):
    """Create a TPU node with gcloud CLI."""
    # TODO(suquark, zhwu): move this to GcpTpuNodeInstance.
    tpu_name = tpu_node_config['name']
    tpu_type = tpu_node_config['acceleratorType']
    try:
        cmd = (f'gcloud compute tpus create {tpu_name} '
               f'--project={project_id} '
               f'--zone={zone} '
               f'--version={tpu_node_config["runtimeVersion"]} '
               f'--accelerator-type={tpu_type} '
               f'--network={vpc_name}')
        logger.debug(f'Creating TPU {tpu_name} with command:\n{cmd}')
        proc = subprocess.run(
            f'yes | {cmd}',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            check=True,
        )
        stdout = proc.stdout.decode('ascii')
        logger.debug(stdout)
    except subprocess.CalledProcessError as e:
        stderr = e.stderr.decode('ascii')
        logger.debug(stderr)
        if 'ALREADY_EXISTS' in stderr:
            # FIXME: should use 'start' on stopped TPUs, replacing
            # 'create'. Or it can be in a "deleting" state. Investigate the
            # right thing to do (force kill + re-provision?).
            logger.warning(f'TPU {tpu_name} already exists; skipped creation.')
            return
        provisioner_err = common.ProvisionerError(TPU_NODE_CREATION_FAILURE)
        if 'RESOURCE_EXHAUSTED' in stderr:
            provisioner_err.errors = [{
                'code': 'RESOURCE_EXHAUSTED',
                'domain': 'tpu',
                'message': f'TPU {tpu_name} creation failed due to quota '
                           'exhaustion. Please visit '
                           'https://console.cloud.google.com/iam-admin/quotas '
                           'for more information.'
            }]
            _format_and_log_message_from_errors(provisioner_err.errors, e, zone)
            raise provisioner_err from e

        if 'PERMISSION_DENIED' in stderr:
            provisioner_err.errors = [{
                'code': 'PERMISSION_DENIED',
                'domain': 'tpu',
                'message': 'TPUs are not available in this zone.'
            }]
            _format_and_log_message_from_errors(provisioner_err.errors, e, zone)
            raise provisioner_err from e

        if 'no more capacity in the zone' in stderr:
            provisioner_err.errors = [{
                'code': 'CapacityExceeded',
                'domain': 'tpu',
                'message': 'No more capacity in this zone.'
            }]
            _format_and_log_message_from_errors(provisioner_err.errors, e, zone)
            raise provisioner_err from e

        if 'CloudTpu received an invalid AcceleratorType' in stderr:
            # INVALID_ARGUMENT: CloudTpu received an invalid
            # AcceleratorType, "v3-8" for zone "us-central1-c". Valid
            # values are "v2-8, ".
            provisioner_err.errors = [{
                'code': 'INVALID_ARGUMENT',
                'domain': 'tpu',
                'message': (f'TPU type {tpu_type} is not available in this '
                            f'zone {zone}.')
            }]
            _format_and_log_message_from_errors(provisioner_err.errors, e, zone)
            raise provisioner_err from e

        # TODO(zhwu): Add more error code handling, if needed.
        provisioner_err.errors = [{
            'code': 'UNKNOWN',
            'domain': 'tpu',
            'message': stderr
        }]
        _format_and_log_message_from_errors(provisioner_err.errors, e, zone)
        raise provisioner_err from e


def delete_tpu_node(project_id: str, zone: str, tpu_node_config: Dict[str,
                                                                      str]):
    """Delete a TPU node with gcloud CLI.

    This is used for both stopping and terminating a cluster with a TPU node. It
    is ok to call this function to delete the TPU node when stopping the cluster
    because the host VM will be stopped and have all the information preserved.
    """
    tpu_name = tpu_node_config['name']
    try:
        cmd = (f'gcloud compute tpus delete {tpu_name} '
               f'--project={project_id} '
               f'--zone={zone}')
        logger.debug(f'Deleting TPU {tpu_name} with cmd:\n{cmd}')
        proc = subprocess.run(
            f'yes | {cmd}',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            check=True,
        )
        stdout = proc.stdout.decode('ascii')
        logger.debug(stdout)
    except subprocess.CalledProcessError as e:
        stdout = e.stdout.decode('ascii')
        stderr = e.stderr.decode('ascii')
        if 'ERROR: (gcloud.compute.tpus.delete) NOT_FOUND' in stderr:
            logger.warning(f'TPU {tpu_name} does not exist; skipped deletion.')
        else:
            raise RuntimeError(f'\nFailed to terminate TPU node {tpu_name} for '
                               'cluster {cluster_name}:\n'
                               '**** STDOUT ****\n'
                               f'{stdout}\n'
                               '**** STDERR ****\n'
                               f'{stderr}') from e



"""Statuses enum for SkyPilot resources."""

import enum

import colorama


class ClusterStatus(enum.Enum):
    """Cluster status as recorded in local cache.

    This can be different from the actual cluster status, and can be refreshed
    by running ``sky status --refresh``.
    """
    # NOTE: these statuses are as recorded in our local cache, the table
    # 'clusters'.  The actual cluster state may be different (e.g., an UP
    # cluster getting killed manually by the user or the cloud provider).

    INIT = 'INIT'
    """Initializing.

    This means a provisioning has started but has not successfully finished. The
    cluster may be undergoing setup, may have failed setup, may be live or down.
    """

    UP = 'UP'
    """The cluster is up. This means a provisioning has previously succeeded."""

    STOPPED = 'STOPPED'
    """The cluster is stopped."""
    PENDING = 'PENDING'
    """The cluster is pending scheduling.

    NOTE: This state is for display only and should not be used in state
    machine logic without necessary considerations.
    """

    def colored_str(self):
        color = _STATUS_TO_COLOR[self]
        return f'{color}{self.value}{colorama.Style.RESET_ALL}'


_STATUS_TO_COLOR = {
    ClusterStatus.INIT: colorama.Fore.BLUE,
    ClusterStatus.UP: colorama.Fore.GREEN,
    ClusterStatus.STOPPED: colorama.Fore.YELLOW,
    ClusterStatus.PENDING: colorama.Fore.CYAN,
}


class StorageStatus(enum.Enum):
    """Storage status as recorded in table 'storage'."""

    # Initializing and uploading storage
    INIT = 'INIT'

    # Initialization failed
    INIT_FAILED = 'INIT_FAILED'

    # Failed to Upload to Cloud
    UPLOAD_FAILED = 'UPLOAD_FAILED'

    # Finished uploading, in terminal state
    READY = 'READY'


class VolumeStatus(enum.Enum):
    """Volume status as recorded in table 'volumes'."""

    # Volume is ready to be used
    READY = 'READY'

    # Volume is being used
    IN_USE = 'IN_USE'



"""Add links column for storing cluster instance links.

Revision ID: 011
Revises: 010
Create Date: 2026-01-07

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '011'
down_revision: Union[str, Sequence[str], None] = '010'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add links column to store instance links as JSON."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('spot',
                                             'links',
                                             sa.JSON(),
                                             server_default=None)


def downgrade():
    """No downgrade logic."""
    pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from sky.schemas.generated import jobsv1_pb2 as sky_dot_schemas_dot_generated_dot_jobsv1__pb2

GRPC_GENERATED_VERSION = '1.63.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in sky/schemas/generated/jobsv1_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class JobsServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddJob = channel.unary_unary(
                '/jobs.v1.JobsService/AddJob',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobResponse.FromString,
                _registered_method=True)
        self.QueueJob = channel.unary_unary(
                '/jobs.v1.JobsService/QueueJob',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobResponse.FromString,
                _registered_method=True)
        self.UpdateStatus = channel.unary_unary(
                '/jobs.v1.JobsService/UpdateStatus',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusResponse.FromString,
                _registered_method=True)
        self.GetJobQueue = channel.unary_unary(
                '/jobs.v1.JobsService/GetJobQueue',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueResponse.FromString,
                _registered_method=True)
        self.CancelJobs = channel.unary_unary(
                '/jobs.v1.JobsService/CancelJobs',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsResponse.FromString,
                _registered_method=True)
        self.FailAllInProgressJobs = channel.unary_unary(
                '/jobs.v1.JobsService/FailAllInProgressJobs',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsResponse.FromString,
                _registered_method=True)
        self.TailLogs = channel.unary_stream(
                '/jobs.v1.JobsService/TailLogs',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsResponse.FromString,
                _registered_method=True)
        self.GetJobStatus = channel.unary_unary(
                '/jobs.v1.JobsService/GetJobStatus',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusResponse.FromString,
                _registered_method=True)
        self.GetJobSubmittedTimestamp = channel.unary_unary(
                '/jobs.v1.JobsService/GetJobSubmittedTimestamp',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampResponse.FromString,
                _registered_method=True)
        self.GetJobEndedTimestamp = channel.unary_unary(
                '/jobs.v1.JobsService/GetJobEndedTimestamp',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampResponse.FromString,
                _registered_method=True)
        self.GetLogDirsForJobs = channel.unary_unary(
                '/jobs.v1.JobsService/GetLogDirsForJobs',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsResponse.FromString,
                _registered_method=True)
        self.GetJobExitCodes = channel.unary_unary(
                '/jobs.v1.JobsService/GetJobExitCodes',
                request_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesResponse.FromString,
                _registered_method=True)


class JobsServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AddJob(self, request, context):
        """Add a new job to the database.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def QueueJob(self, request, context):
        """Queue a job for execution.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateStatus(self, request, context):
        """Update status of all jobs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobQueue(self, request, context):
        """Get the job queue.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelJobs(self, request, context):
        """Cancel jobs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FailAllInProgressJobs(self, request, context):
        """Fail all in-progress jobs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TailLogs(self, request, context):
        """Tail job logs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobStatus(self, request, context):
        """Get job status.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobSubmittedTimestamp(self, request, context):
        """Get job submitted timestamp.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobEndedTimestamp(self, request, context):
        """Get job ended timestamp.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLogDirsForJobs(self, request, context):
        """Get log directories for jobs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobExitCodes(self, request, context):
        """Get job exit codes.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_JobsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'AddJob': grpc.unary_unary_rpc_method_handler(
                    servicer.AddJob,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobResponse.SerializeToString,
            ),
            'QueueJob': grpc.unary_unary_rpc_method_handler(
                    servicer.QueueJob,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobResponse.SerializeToString,
            ),
            'UpdateStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateStatus,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusResponse.SerializeToString,
            ),
            'GetJobQueue': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobQueue,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueResponse.SerializeToString,
            ),
            'CancelJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelJobs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsResponse.SerializeToString,
            ),
            'FailAllInProgressJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.FailAllInProgressJobs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsResponse.SerializeToString,
            ),
            'TailLogs': grpc.unary_stream_rpc_method_handler(
                    servicer.TailLogs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsResponse.SerializeToString,
            ),
            'GetJobStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobStatus,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusResponse.SerializeToString,
            ),
            'GetJobSubmittedTimestamp': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobSubmittedTimestamp,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampResponse.SerializeToString,
            ),
            'GetJobEndedTimestamp': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobEndedTimestamp,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampResponse.SerializeToString,
            ),
            'GetLogDirsForJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLogDirsForJobs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsResponse.SerializeToString,
            ),
            'GetJobExitCodes': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobExitCodes,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'jobs.v1.JobsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class JobsService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AddJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/AddJob',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.AddJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def QueueJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/QueueJob',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.QueueJobResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdateStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/UpdateStatus',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.UpdateStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobQueue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetJobQueue',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobQueueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/CancelJobs',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.CancelJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FailAllInProgressJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/FailAllInProgressJobs',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.FailAllInProgressJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TailLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/jobs.v1.JobsService/TailLogs',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.TailLogsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetJobStatus',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobSubmittedTimestamp(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetJobSubmittedTimestamp',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobSubmittedTimestampResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobEndedTimestamp(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetJobEndedTimestamp',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobEndedTimestampResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLogDirsForJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetLogDirsForJobs',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetLogDirsForJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobExitCodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/jobs.v1.JobsService/GetJobExitCodes',
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_jobsv1__pb2.GetJobExitCodesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Utility functions for threads."""

import threading
from typing import Any, Dict, Generic, Optional, overload, TypeVar

from sky.utils import common_utils


class SafeThread(threading.Thread):
    """A thread that can catch exceptions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._exc = None

    def run(self):
        try:
            super().run()
        except BaseException as e:  # pylint: disable=broad-except
            self._exc = e

    @property
    def format_exc(self) -> Optional[str]:
        if self._exc is None:
            return None
        return common_utils.format_exception(self._exc)


# pylint: disable=invalid-name
KeyType = TypeVar('KeyType')
ValueType = TypeVar('ValueType')


# Google style guide: Do not rely on the atomicity of built-in types.
# Our launch and down process pool will be used by multiple threads,
# therefore we need to use a thread-safe dict.
# see https://google.github.io/styleguide/pyguide.html#218-threading
class ThreadSafeDict(Generic[KeyType, ValueType]):
    """A thread-safe dict."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._dict: Dict[KeyType, ValueType] = dict(*args, **kwargs)
        self._lock = threading.Lock()

    def __getitem__(self, key: KeyType) -> ValueType:
        with self._lock:
            return self._dict.__getitem__(key)

    def __setitem__(self, key: KeyType, value: ValueType) -> None:
        with self._lock:
            return self._dict.__setitem__(key, value)

    def __delitem__(self, key: KeyType) -> None:
        with self._lock:
            return self._dict.__delitem__(key)

    def __len__(self) -> int:
        with self._lock:
            return self._dict.__len__()

    def __contains__(self, key: KeyType) -> bool:
        with self._lock:
            return self._dict.__contains__(key)

    def items(self):
        with self._lock:
            return self._dict.items()

    def values(self):
        with self._lock:
            return self._dict.values()

    @overload
    def get(self, key: KeyType, default: ValueType) -> ValueType:
        ...

    @overload
    def get(self,
            key: KeyType,
            default: Optional[ValueType] = None) -> Optional[ValueType]:
        ...

    def get(self,
            key: KeyType,
            default: Optional[ValueType] = None) -> Optional[ValueType]:
        with self._lock:
            return self._dict.get(key, default)

    def pop(self, key: KeyType) -> Optional[ValueType]:
        with self._lock:
            return self._dict.pop(key, None)



"""Persistent KV cache, backed by a sqlite or postgres database."""
import functools
import threading
import time
from typing import Optional

import sqlalchemy
from sqlalchemy import exc as sqlalchemy_exc
from sqlalchemy import orm
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects import sqlite
from sqlalchemy.ext import declarative

from sky import sky_logging
from sky.metrics import utils as metrics_lib
from sky.utils import common_utils
from sky.utils.db import db_utils
from sky.utils.db import migration_utils

logger = sky_logging.init_logger(__name__)

_SQLALCHEMY_ENGINE: Optional[sqlalchemy.engine.Engine] = None
_SQLALCHEMY_ENGINE_LOCK = threading.Lock()

Base = declarative.declarative_base()

kv_cache_table = sqlalchemy.Table(
    'kv_cache',
    Base.metadata,
    sqlalchemy.Column('key', sqlalchemy.Text, primary_key=True),
    sqlalchemy.Column('value', sqlalchemy.Text),
    sqlalchemy.Column('expires_at', sqlalchemy.Float),
)


def create_table(engine: sqlalchemy.engine.Engine):
    # Enable WAL mode to avoid locking issues.
    # See: issue #1441 and PR #1509
    # https://github.com/microsoft/WSL/issues/2395
    # TODO(romilb): We do not enable WAL for WSL because of known issue in WSL.
    #  This may cause the database locked problem from WSL issue #1441.
    if (engine.dialect.name == db_utils.SQLAlchemyDialect.SQLITE.value and
            not common_utils.is_wsl()):
        try:
            with orm.Session(engine) as session:
                session.execute(sqlalchemy.text('PRAGMA journal_mode=WAL'))
                session.commit()
        except sqlalchemy_exc.OperationalError as e:
            if 'database is locked' not in str(e):
                raise
            # If the database is locked, it is OK to continue, as the WAL mode
            # is not critical and is likely to be enabled by other processes.

    migration_utils.safe_alembic_upgrade(engine,
                                         migration_utils.KV_CACHE_DB_NAME,
                                         migration_utils.KV_CACHE_VERSION)


# We wrap the sqlalchemy engine initialization in a thread
# lock to ensure that multiple threads do not initialize the
# engine which could result in a rare race condition where
# a session has already been created with _SQLALCHEMY_ENGINE = e1,
# and then another thread overwrites _SQLALCHEMY_ENGINE = e2
# which could result in e1 being garbage collected unexpectedly.
def initialize_and_get_db() -> sqlalchemy.engine.Engine:
    global _SQLALCHEMY_ENGINE

    if _SQLALCHEMY_ENGINE is not None:
        return _SQLALCHEMY_ENGINE
    with _SQLALCHEMY_ENGINE_LOCK:
        if _SQLALCHEMY_ENGINE is not None:
            return _SQLALCHEMY_ENGINE
        # get an engine to the db
        engine = db_utils.get_engine('kv_cache')

        # run migrations if needed
        create_table(engine)
        _SQLALCHEMY_ENGINE = engine
        return _SQLALCHEMY_ENGINE


def _init_db(func):
    """Initialize the database."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        initialize_and_get_db()
        return func(*args, **kwargs)

    return wrapper


initialize_and_get_db()


@_init_db
@metrics_lib.time_me
def add_or_update_cache_entry(
    key: str,
    value: str,
    expires_at: float,
) -> None:
    """Store the mapping from user hash to user name for display purposes.

    Args:
        key: The key of the cache entry.
        value: The value of the cache entry.
        expires_at: The timestamp when the cache entry expires.
    """
    assert _SQLALCHEMY_ENGINE is not None
    if (_SQLALCHEMY_ENGINE.dialect.name ==
            db_utils.SQLAlchemyDialect.SQLITE.value):
        insert_func = sqlite.insert
    elif (_SQLALCHEMY_ENGINE.dialect.name ==
          db_utils.SQLAlchemyDialect.POSTGRESQL.value):
        insert_func = postgresql.insert
    else:
        raise ValueError('Unsupported database dialect')

    with orm.Session(_SQLALCHEMY_ENGINE) as session:
        insert_stmt = insert_func(kv_cache_table).values(key=key,
                                                         value=value,
                                                         expires_at=expires_at)
        do_update_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[kv_cache_table.c.key],
            set_={
                kv_cache_table.c.value: value,
                kv_cache_table.c.expires_at: expires_at
            })
        session.execute(do_update_stmt)

        session.commit()


@_init_db
@metrics_lib.time_me
def get_cache_entry(key: str) -> Optional[str]:
    """Get the value of the cache entry.

    Args:
        key: The key of the cache entry.
    """
    assert _SQLALCHEMY_ENGINE is not None
    with orm.Session(_SQLALCHEMY_ENGINE) as session:
        result = session.execute(
            sqlalchemy.select(kv_cache_table.c.value).where(
                kv_cache_table.c.key == key).where(
                    kv_cache_table.c.expires_at > time.time()))
        return result.scalar()



"""Utilities to setup SSH Tunnel"""
import os
import random
import re
import subprocess
import sys
from typing import Set

import colorama

from sky import sky_logging
from sky.ssh_node_pools import constants
from sky.ssh_node_pools.deploy import utils as deploy_utils

logger = sky_logging.init_logger(__name__)

# Get the directory of this script
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


def _get_used_localhost_ports() -> Set[int]:
    """Get SSH port forwardings already in use on localhost"""
    used_ports = set()

    # Get ports from netstat (works on macOS and Linux)
    try:
        if sys.platform == 'darwin':
            # macOS
            result = subprocess.run(['netstat', '-an', '-p', 'tcp'],
                                    capture_output=True,
                                    text=True,
                                    check=False)
        else:
            # Linux and other Unix-like systems
            result = subprocess.run(['netstat', '-tln'],
                                    capture_output=True,
                                    text=True,
                                    check=False)

        if result.returncode == 0:
            # Look for lines with 'localhost:<port>' or '127.0.0.1:<port>'
            for line in result.stdout.splitlines():
                if '127.0.0.1:' in line or 'localhost:' in line:
                    match = re.search(r':(64\d\d)\s', line)
                    if match:
                        port = int(match.group(1))
                        if 6400 <= port <= 6500:  # Only consider our range
                            used_ports.add(port)
    except (subprocess.SubprocessError, FileNotFoundError):
        # If netstat fails, try another approach
        pass

    # Also check ports from existing kubeconfig entries
    try:
        result = subprocess.run([
            'kubectl', 'config', 'view', '-o',
            'jsonpath=\'{.clusters[*].cluster.server}\''
        ],
                                capture_output=True,
                                text=True,
                                check=False)

        if result.returncode == 0:
            # Look for localhost URLs with ports
            for url in result.stdout.split():
                if 'localhost:' in url or '127.0.0.1:' in url:
                    match = re.search(r':(\d+)', url)
                    if match:
                        port = int(match.group(1))
                        if 6400 <= port <= 6500:  # Only consider our range
                            used_ports.add(port)
    except subprocess.SubprocessError:
        pass

    return used_ports


def get_available_port(start: int = 6443, end: int = 6499) -> int:
    """Get an available port in the given range not used by other tunnels"""
    used_ports = _get_used_localhost_ports()

    # Try to use port 6443 first if available for the first cluster
    if start == 6443 and start not in used_ports:
        return start

    # Otherwise find any available port in the range
    available_ports = list(set(range(start, end + 1)) - used_ports)

    if not available_ports:
        # If all ports are used, pick a random one from our range
        # (we'll terminate any existing connection in the setup)
        return random.randint(start, end)

    # Sort to get deterministic allocation
    available_ports.sort()
    return available_ports[0]


def setup_kubectl_ssh_tunnel(head_node,
                             ssh_user,
                             ssh_key,
                             context_name,
                             use_ssh_config=False):
    """Set up kubeconfig exec credential plugin for SSH tunnel"""
    logger.info(f'{colorama.Fore.YELLOW}➜ Setting up SSH tunnel for '
                f'Kubernetes API access...{colorama.Style.RESET_ALL}')

    # Get an available port for this cluster
    port = get_available_port()

    # Paths to scripts
    tunnel_script = os.path.join(SCRIPT_DIR, 'tunnel', 'ssh-tunnel.sh')

    # Make sure scripts are executable
    os.chmod(tunnel_script, 0o755)

    # Certificate files
    client_cert_file = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                    f'{context_name}-cert.pem')
    client_key_file = os.path.join(constants.NODE_POOLS_INFO_DIR,
                                   f'{context_name}-key.pem')

    # Update kubeconfig to use localhost with the selected port
    deploy_utils.run_command([
        'kubectl', 'config', 'set-cluster', context_name,
        f'--server=https://127.0.0.1:{port}', '--insecure-skip-tls-verify=true'
    ])

    # Build the exec args list based on auth method
    exec_args = [
        '--exec-command', tunnel_script, '--exec-api-version',
        'client.authentication.k8s.io/v1beta1'
    ]

    # Set credential TTL to force frequent tunnel checks
    ttl_seconds = 30

    # Verify if we have extracted certificate data files
    has_cert_files = os.path.isfile(client_cert_file) and os.path.isfile(
        client_key_file)
    if has_cert_files:
        logger.info(f'{colorama.Fore.GREEN}Client certificate data extracted '
                    'and will be used for authentication'
                    f'{colorama.Style.RESET_ALL}')

    if use_ssh_config:
        deploy_utils.run_command(
            ['kubectl', 'config', 'set-credentials', context_name] + exec_args +
            [
                '--exec-arg=--context', f'--exec-arg={context_name}',
                '--exec-arg=--port', f'--exec-arg={port}', '--exec-arg=--ttl',
                f'--exec-arg={ttl_seconds}', '--exec-arg=--use-ssh-config',
                '--exec-arg=--host', f'--exec-arg={head_node}'
            ])
    else:
        deploy_utils.run_command(
            ['kubectl', 'config', 'set-credentials', context_name] + exec_args +
            [
                '--exec-arg=--context', f'--exec-arg={context_name}',
                '--exec-arg=--port', f'--exec-arg={port}', '--exec-arg=--ttl',
                f'--exec-arg={ttl_seconds}', '--exec-arg=--host',
                f'--exec-arg={head_node}', '--exec-arg=--user',
                f'--exec-arg={ssh_user}', '--exec-arg=--ssh-key',
                f'--exec-arg={ssh_key}'
            ])

    logger.info(f'{colorama.Fore.GREEN}✔ SSH tunnel configured through '
                'kubectl credential plugin on port '
                f'{port}{colorama.Style.RESET_ALL}')
    logger.info('Your kubectl connection is now tunneled through SSH '
                f'(port {port}).')
    logger.info('This tunnel will be automatically established when needed.')
    logger.info(f'Credential TTL set to {ttl_seconds}s to ensure tunnel '
                'health is checked frequently.')
    return port


def cleanup_kubectl_ssh_tunnel(cluster_name, context_name):
    """Clean up the SSH tunnel for a specific context"""
    logger.info(f'{colorama.Fore.YELLOW}➜ Cleaning up SSH tunnel for '
                f'`{cluster_name}`...{colorama.Style.RESET_ALL}')

    # Path to cleanup script
    cleanup_script = os.path.join(SCRIPT_DIR, 'tunnel', 'cleanup-tunnel.sh')

    # Make sure script is executable
    if os.path.exists(cleanup_script):
        os.chmod(cleanup_script, 0o755)

        # Run the cleanup script
        subprocess.run([cleanup_script, context_name],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL,
                       check=False)
        logger.info(f'{colorama.Fore.GREEN}✔ SSH tunnel for `{cluster_name}` '
                    f'cleaned up.{colorama.Style.RESET_ALL}')
    else:
        logger.error(f'{colorama.Fore.YELLOW}Cleanup script not found: '
                     f'{cleanup_script}{colorama.Style.RESET_ALL}')



"""vsphere utils
"""
import http.cookies as http_cookies
import os
import ssl
from typing import Any, Dict, List, Optional

from sky import exceptions
from sky import sky_logging
from sky.adaptors import vsphere as vsphere_adaptor
from sky.catalog import vsphere_catalog
from sky.catalog.common import get_catalog_path
from sky.catalog.data_fetchers.fetch_vsphere import initialize_accelerators_csv
from sky.catalog.data_fetchers.fetch_vsphere import initialize_hosts_csv
from sky.catalog.data_fetchers.fetch_vsphere import initialize_images_csv
from sky.catalog.data_fetchers.fetch_vsphere import (
    initialize_instance_image_mapping_csv)
from sky.catalog.data_fetchers.fetch_vsphere import initialize_vms_csv
from sky.provision.vsphere.common import vim_utils
from sky.provision.vsphere.common.cls_api_client import ClsApiClient
from sky.provision.vsphere.common.cls_api_helper import ClsApiHelper
from sky.provision.vsphere.common.id_generator import generate_random_uuid
from sky.provision.vsphere.common.service_manager_factory import (
    ServiceManagerFactory)
from sky.provision.vsphere.common.vim_utils import create_spec_with_script
from sky.provision.vsphere.common.vim_utils import poweron_vm
from sky.provision.vsphere.common.vim_utils import wait_for_tasks
from sky.provision.vsphere.common.vim_utils import wait_internal_ip_ready
from sky.utils import yaml_utils

logger = sky_logging.init_logger(__name__)

CREDENTIALS_PATH = '~/.vsphere/credential.yaml'
VM_PREFIX = 'skypilot-vm-'


class VsphereError(Exception):
    pass


class VsphereClient:
    """VsphereClient
    """

    def __init__(
        self,
        server: str,
        username: str,
        password: str,
        clusters: List[Dict[str, str]],
        skip_verification: bool,
    ) -> None:
        self.__server = server
        self.__username = username
        self.__password = password
        self.clusters = clusters
        self.skip_verification = skip_verification
        self.servicemanager = None
        self.tag = None
        self.tag_association = None

    def connect(self):
        if not self.servicemanager:
            self.servicemanager = ServiceManagerFactory.get_service_manager(
                self.__server, self.__username, self.__password,
                self.skip_verification)

    def disconnect(self):
        if self.servicemanager:
            self.servicemanager.disconnect()

    def remove_instances(self, *instance_ids: str) -> Dict[str, Any]:  # pylint: disable=unused-argument
        """Terminate instances."""
        ### TODO: Add implementation
        # pass
        return {}

    def list_instances(self):
        ### TODO: Add implementation
        return []

    def filter_instances(
        self,
        instance_uuids: Optional[List[str]] = None,
        filters: Optional[List[Dict[str, str]]] = None,
        vsphere_cluster_names: Optional[List[str]] = None,
    ):
        """Filter the vm instances using custom attribute
        """
        instances: List[Any] = []
        service_manager = self.servicemanager
        # If no instance_ids passed in, then get all VMs in vsphere cluster;
        # Else get instances by instance_ids.
        if service_manager:
            if not instance_uuids:
                instances = vim_utils.filter_vms_by_cluster_name(
                    service_manager.content, vsphere_cluster_names)
            else:
                for inst_id in instance_uuids:
                    instance_obj = (
                        service_manager.content.searchIndex.FindByUuid(
                            None, inst_id, True, True))
                    if instance_obj:
                        instances.append(instance_obj)

            if not filters:
                return instances

            filtered_result = []
            filter_keys = [f['Key'] for f in filters]
            for inst in instances:
                #  If the param: instance_ids is empty or none, the
                #  [startswith(VM_PREFIX)] condition
                #  will quickly filter out instances that are not part
                #  of skypilot
                if inst.name.startswith(VM_PREFIX):
                    # Filter the instances using Custom Attributes
                    cust_attributes = [(f.name, v.value)
                                       for f in inst.availableField
                                       if f.name in filter_keys
                                       for v in inst.customValue
                                       if f.key == v.key]
                    for filter_item in filters:
                        if (filter_item['Key'],
                                filter_item['Value']) not in cust_attributes:
                            break
                    else:
                        filtered_result.append(inst)

            return filtered_result
        else:
            raise VsphereError('Failed to connect to vSphere.')

    def check_credential(self):
        self.connect()

    def create_instances(
        self,
        cluster: str,
        host_mobid: str,
        lib_item_id: str,
        spec,
        customization_spec_str: str,
    ):
        # Set up
        self.vm_name = VM_PREFIX + str(generate_random_uuid())
        self.connect()
        self.client = ClsApiClient(self.servicemanager)
        self.helper = ClsApiHelper(self.client, self.skip_verification)
        service_manager = self.servicemanager
        if service_manager:
            # Find the cluster's resource pool moid
            cluster_obj = vim_utils.get_objs_by_names(
                service_manager.content,
                [vsphere_adaptor.get_vim().ClusterComputeResource], [cluster])
            if not cluster_obj.get(cluster):
                logger.error(f'Failed to find cluster {cluster}')
                raise exceptions.ResourcesUnavailableError(
                    f'Failed to find cluster {cluster}')
            cluster_obj = cluster_obj[cluster]
            logger.info('Cluster Moref: {0}'.format(cluster_obj))

            # Find the deployment target
            deployment_target = vsphere_adaptor.get_ovf_client(
            ).LibraryItem.DeploymentTarget(
                resource_pool_id=cluster_obj.resourcePool._GetMoId(),  # pylint: disable=protected-access
                host_id=host_mobid)

            ovf_summary = self.client.ovf_lib_item_service.filter(
                ovf_library_item_id=lib_item_id, target=deployment_target)
            logger.info('Found an OVF template :{0} to deploy.'.format(
                ovf_summary.name))

            # Find out the storage profile id
            profile_id = self.get_skypilot_profile_id()
            if profile_id is None:
                logger.error('The VM storage policy \'skypilot_policy\''
                             ' is not available.')
                raise exceptions.ResourcesUnavailableError(
                    'The VM storage policy \'skypilot_policy\''
                    ' is not available.')
            # Deploy the ovf template
            vm_obj = self.helper.deploy_ovf_template(
                vm_name=self.vm_name,
                lib_item_id=lib_item_id,
                ovf_summary=ovf_summary,
                deployment_target=deployment_target,
                storage_profile_id=profile_id,
            )

            customization_spec = create_spec_with_script(
                vm_obj, customization_spec_str)
            check_customization_spec = vm_obj.CheckCustomizationSpec(
                spec=customization_spec)
            if check_customization_spec:
                logger.error(
                    f'Check customization spec failed for VM {vm_obj.name}: '
                    f'{check_customization_spec}')
                raise exceptions.ResourcesUnavailableError(
                    f'Check customization spec failed for VM {vm_obj.name}: '
                    f'{check_customization_spec}')
            wait_for_tasks(
                self.client.service_manager.content,
                [
                    vm_obj.Reconfigure(spec),
                    vm_obj.CustomizeVM_Task(spec=customization_spec),
                ],
            )
            poweron_vm(self.client.service_manager.content, vm_obj)
            wait_internal_ip_ready(vm_obj)
            return vm_obj.summary.config.instanceUuid
        else:
            raise VsphereError('Failed to connect to vSphere.')

    def set_tags(self, head_instance_uuid: str, tags: List[Dict[str, str]]):
        """Set tag for vm instance"""
        service_manager = self.servicemanager
        if service_manager:
            content = self.servicemanager.content
            cfm = content.customFieldsManager

            # Get the instance Object by instance_uuid
            vm = content.searchIndex.FindByUuid(None, head_instance_uuid, True,
                                                True)

            # Create the Field if not exsit, then set value.
            fields = [f.name for f in vm.availableField]
            for tag in tags:
                if tag['Key'] not in fields:
                    new_field = cfm.AddFieldDefinition(
                        tag['Key'],
                        vsphere_adaptor.get_vim().VirtualMachine)
                    cfm.SetField(vm, new_field.key, tag['Value'])
                    continue
                exi_key = [
                    fld.key
                    for fld in vm.availableField
                    if fld.name == tag['Key']
                ][0]
                cfm.SetField(vm, exi_key, tag['Value'])
        else:
            raise VsphereError('Failed to connect to vSphere.')

    def get_skypilot_profile_id(self):
        pm = self.get_pbm_manager()
        profile_ids = pm.PbmQueryProfile(
            resourceType=vsphere_adaptor.get_pbm().profile.ResourceType(
                resourceType='STORAGE'),
            profileCategory='REQUIREMENT',
        )

        # hard code here. should support configure later.
        profile_name = 'skypilot_policy'
        storage_profile_id = None
        if profile_ids:
            profiles = pm.PbmRetrieveContent(profileIds=profile_ids)
            for profile in profiles:
                if profile_name in profile.name:
                    storage_profile_id = profile.profileId.uniqueId
                    break
        return storage_profile_id

    def get_pbm_manager(self):
        self.connect()
        if self.servicemanager is None:
            raise VsphereError('Failed to connect to vSphere.')
        if self.servicemanager.si is None:
            raise VsphereError('Failed to connect to vSphere.')
        pbm_si, pm_content = self._create_pbm_connection(  # pylint: disable=unused-variable
            self.servicemanager.si._stub)  # pylint: disable=protected-access
        pm = pm_content.profileManager
        return pm

    # we should not call this method directly
    def _create_pbm_connection(self, vpxd_stub):
        session_cookie = vpxd_stub.cookie.split('\"')[1]
        http_context = vsphere_adaptor.get_vmomi_support().GetHttpContext()
        cookie = http_cookies.SimpleCookie()
        cookie['vmware_soap_session'] = session_cookie
        http_context['cookies'] = cookie
        vsphere_adaptor.get_vmomi_support().GetRequestContext(
        )['vcSessionCookie'] = session_cookie
        hostname = vpxd_stub.host.split(':')[0]

        context = None
        if hasattr(ssl, '_create_unverified_context'):
            context = ssl._create_unverified_context()  # pylint: disable=protected-access
        pbm_stub = vsphere_adaptor.get_pyvmomi().SoapStubAdapter(
            host=hostname,
            version='pbm.version.version1',
            path='/pbm/sdk',
            poolSize=0,
            sslContext=context,
        )
        pbm_si = vsphere_adaptor.get_pbm().ServiceInstance(
            'ServiceInstance', pbm_stub)
        pbm_content = pbm_si.RetrieveContent()

        return pbm_si, pbm_content


def get_vsphere_credentials(name=None):
    """The credential format is:
            vcenters:
              - name: vcenter1
                username: xxxx
                password: xxxx
                skip_verification: true
                clusters:
                - name: cluster1
                - name: cluster2
              - name: vcenter2
                username: xxxx
                password: xxxx
                skip_verification: true
                clusters:
                - name: cluster1
                - name: cluster2
    """
    credential_path = os.path.expanduser(CREDENTIALS_PATH)
    assert os.path.exists(
        credential_path), f'Missing credential file at {credential_path}.'
    with open(credential_path, 'r', encoding='utf-8') as file:
        credential = yaml_utils.safe_load(file)
        vcenters = credential['vcenters']
        if name is None:
            return vcenters
        for vcenter in vcenters:
            if vcenter['name'] == name:
                return vcenter
        raise VsphereError(f' Failed to find vcenter which name is: {name}.')


def initialize_vsphere_data():
    os.makedirs(get_catalog_path('vsphere'), exist_ok=True)
    initialize_accelerators_csv()

    vms_csv_path = get_catalog_path('vsphere/vms.csv')
    with open(vms_csv_path, 'w', encoding='utf-8') as f:
        f.write(vsphere_catalog.VSPHERE_CATALOG_HEADER + '\n')
    images_csv_path = get_catalog_path('vsphere/images.csv')
    with open(images_csv_path, 'w', encoding='utf-8') as f:
        f.write('ImageID,vCenter,CPU,Memory,OS,OSVersion,GpuTags\n')
    hosts_csv_path = get_catalog_path('vsphere/hosts.csv')
    with open(hosts_csv_path, 'w', encoding='utf-8') as f:
        f.write(
            'HostName,MobID,vCenter,Datacenter,Cluster,TotalCPUs,AvailableCPUs,'
            'TotalMemory(MB),AvailableMemory(MB),GPU,cpuMhz,UUID\n')
    instance_image_mapping_csv_path = get_catalog_path(
        'vsphere/instance_image_mapping.csv')
    with open(instance_image_mapping_csv_path, 'w', encoding='utf-8') as f:
        f.write('InstanceType,ImageID,vCenter\n')
    skip_key = 'skip_verification'
    for vcenter in get_vsphere_credentials():
        if skip_key not in vcenter:
            vcenter[skip_key] = False
        vc_object = VsphereClient(
            vcenter['name'],
            vcenter['username'],
            vcenter['password'],
            vcenter['clusters'],
            vcenter['skip_verification'],
        )
        vcenter_name = vcenter['name']
        vc_object.connect()
        vc_servicemanager = vc_object.servicemanager
        if vc_servicemanager is None or vc_servicemanager.content is None:
            raise VsphereError('Failed to connect to vSphere.')
        vc_content = vc_servicemanager.content

        cluster_name_dicts = vc_object.clusters
        hosts = vim_utils.get_hosts_by_cluster_names(vc_content, vcenter_name,
                                                     cluster_name_dicts)
        initialize_hosts_csv(hosts_csv_path, hosts)
        initialize_vms_csv(vms_csv_path, hosts, vcenter_name)
        initialize_images_csv(images_csv_path, vc_object, vcenter_name)
        initialize_instance_image_mapping_csv(vms_csv_path, images_csv_path,
                                              instance_image_mapping_csv_path)
        if vc_object.servicemanager is not None:
            vc_object.servicemanager.disconnect()



"""SSL Helper
"""
import ssl
import typing

from sky.adaptors import common as adaptors_common

if typing.TYPE_CHECKING:
    import requests
else:
    requests = adaptors_common.LazyImport('requests')


def get_unverified_context():
    """    Get an unverified ssl context. Used to disable the server certificate
    verification.
    @return: unverified ssl context.
    """
    context = None
    if hasattr(ssl, '_create_unverified_context'):
        context = ssl._create_unverified_context()  # pylint: disable=protected-access
    return context


def get_unverified_session():
    """    Get a requests session with cert verification disabled.
    Also disable the insecure warnings message.
    Note this is not recommended in production code.
    @return: a requests session with verification disabled.
    """
    session = requests.session()
    session.verify = False
    requests.packages.urllib3.disable_warnings()  # type: ignore[attr-defined]
    return session



"""Version-aware serializers for request return values.

These serializers run at encode() time when remote_api_version is available,
to handle backward compatibility for old clients.

The existing encoders.py handles object -> dict conversion at set_return_value()
time. This module handles dict -> JSON string serialization at encode() time,
with version-aware field filtering for backward compatibility.
"""
from typing import Any, Callable, Dict

import orjson

from sky.server import constants as server_constants
from sky.server import versions

handlers: Dict[str, Callable[[Any], str]] = {}


def register_serializer(*names: str):
    """Decorator to register a version-aware serializer."""

    def decorator(func):
        for name in names:
            if name != server_constants.DEFAULT_HANDLER_NAME:
                name = server_constants.REQUEST_NAME_PREFIX + name
            if name in handlers:
                raise ValueError(f'Serializer {name} already registered: '
                                 f'{handlers[name]}')
            handlers[name] = func
        return func

    return decorator


def get_serializer(name: str) -> Callable[[Any], str]:
    """Get the serializer for a request name."""
    return handlers.get(name, handlers[server_constants.DEFAULT_HANDLER_NAME])


@register_serializer(server_constants.DEFAULT_HANDLER_NAME)
def default_serializer(return_value: Any) -> str:
    """The default serializer."""
    return orjson.dumps(return_value).decode('utf-8')


@register_serializer('kubernetes_node_info')
def serialize_kubernetes_node_info(return_value: Dict[str, Any]) -> str:
    """Serialize kubernetes node info with version compatibility.

    The is_ready field was added in API version 25. Remove it for old clients
    that don't recognize it.
    The cpu_count, memory_gb, cpu_free, and memory_free_gb fields were added
    in API version 26. Remove them for old clients that don't recognize them.
    """
    remote_api_version = versions.get_remote_api_version()
    if (return_value and remote_api_version is not None):
        for node_info in return_value.get('node_info_dict', {}).values():
            if remote_api_version < 25:
                # Remove is_ready field for old clients that don't recognize it
                node_info.pop('is_ready', None)
            if remote_api_version < 26:
                # Remove cpu_count, memory_gb, cpu_free, and
                # memory_free_gb fields for old clients that don't
                # recognize them
                node_info.pop('cpu_count', None)
                node_info.pop('memory_gb', None)
                node_info.pop('cpu_free', None)
                node_info.pop('memory_free_gb', None)
    return orjson.dumps(return_value).decode('utf-8')



"""Workspace management core."""

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Tuple

import filelock

from sky import check as sky_check
from sky import exceptions
from sky import models
from sky import sky_logging
from sky import skypilot_config
from sky.backends import backend_utils
from sky.skylet import constants
from sky.usage import usage_lib
from sky.users import permission
from sky.users import rbac
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import config_utils
from sky.utils import locks
from sky.utils import resource_checker
from sky.utils import schemas
from sky.workspaces import utils as workspaces_utils

logger = sky_logging.init_logger(__name__)

# Lock for workspace configuration updates to prevent race conditions
_WORKSPACE_CONFIG_LOCK_TIMEOUT_SECONDS = 60


@dataclass
class WorkspaceConfigComparison:
    """Result of comparing current and new workspace configurations.

    This class encapsulates the results of analyzing differences between
    workspace configurations, particularly focusing on user access changes
    and their implications for resource validation.

    Attributes:
        only_user_access_changes: True if only allowed_users or private changed
        private_changed: True if private setting changed
        private_old: Old private setting value
        private_new: New private setting value
        allowed_users_changed: True if allowed_users changed
        allowed_users_old: Old allowed users list
        allowed_users_new: New allowed users list
        removed_users: Users removed from allowed_users
        added_users: Users added to allowed_users
    """
    only_user_access_changes: bool
    private_changed: bool
    private_old: bool
    private_new: bool
    allowed_users_changed: bool
    allowed_users_old: List[str]
    allowed_users_new: List[str]
    removed_users: List[str]
    added_users: List[str]


# =========================
# = Workspace Management =
# =========================


def get_workspaces() -> Dict[str, Any]:
    """Returns the workspace config."""
    return workspaces_for_user(common_utils.get_current_user().id)


def _update_workspaces_config(
        workspace_modifier_fn: Callable[[Dict[str, Any]],
                                        None]) -> Dict[str, Any]:
    """Update the workspaces configuration in the config file.

    This function uses file locking to prevent race conditions when multiple
    processes try to update the workspace configuration simultaneously.

    Args:
        workspace_modifier_fn: A function that takes the current workspaces
            dict and modifies it in-place. This ensures all read-modify-write
            operations happen atomically inside the lock.

    Returns:
        The updated workspaces configuration.
    """
    lock_path = skypilot_config.get_skypilot_config_lock_path()
    try:
        with filelock.FileLock(lock_path,
                               _WORKSPACE_CONFIG_LOCK_TIMEOUT_SECONDS):
            # Read the current config inside the lock to ensure we have
            # the latest state
            current_config = skypilot_config.to_dict()
            current_workspaces = current_config.get('workspaces', {}).copy()

            # Apply the modification inside the lock
            workspace_modifier_fn(current_workspaces)

            # Update the config with the modified workspaces
            current_config['workspaces'] = current_workspaces

            # Write the configuration back to the file
            skypilot_config.update_api_server_config_no_lock(current_config)

            return current_workspaces
    except filelock.Timeout as e:
        raise RuntimeError(
            f'Failed to update workspace configuration due to a timeout '
            f'when trying to acquire the lock at {lock_path}. This may '
            'indicate another SkyPilot process is currently updating the '
            'configuration. Please try again or manually remove the lock '
            f'file if you believe it is stale.') from e


def _validate_workspace_config(workspace_name: str,
                               workspace_config: Dict[str, Any]) -> None:
    """Validate the workspace configuration.
    """
    workspace_schema = schemas.get_config_schema(
    )['properties']['workspaces']['additionalProperties']
    try:
        common_utils.validate_schema(
            workspace_config, workspace_schema,
            f'Invalid configuration for workspace {workspace_name!r}: ')
    except exceptions.InvalidSkyPilotConfigError as e:
        # We need to replace this exception with a ValueError because: a) it is
        # more user-friendly and b) it will not be caught by the try-except by
        # the caller the may cause confusion.
        raise ValueError(str(e)) from e


def _compare_workspace_configs(
    current_config: Dict[str, Any],
    new_config: Dict[str, Any],
) -> WorkspaceConfigComparison:
    """Compare current and new workspace configurations.

    Args:
        current_config: The current workspace configuration.
        new_config: The new workspace configuration.

    Returns:
        WorkspaceConfigComparison object containing the comparison results.
    """
    # Get private settings
    private_old = current_config.get('private', False)
    private_new = new_config.get('private', False)
    private_changed = private_old != private_new

    admin_user_ids = permission.permission_service.get_users_for_role(
        rbac.RoleName.ADMIN.value)
    # Get allowed users (resolve to user IDs for comparison)
    allowed_users_old = workspaces_utils.get_workspace_users(
        current_config) if private_old else []
    allowed_users_old += admin_user_ids
    allowed_users_new = workspaces_utils.get_workspace_users(
        new_config) if private_new else []
    allowed_users_new += admin_user_ids

    # Convert to sets for easier comparison
    old_users_set = set(allowed_users_old)
    new_users_set = set(allowed_users_new)

    allowed_users_changed = old_users_set != new_users_set
    removed_users = list(old_users_set - new_users_set)
    added_users = list(new_users_set - old_users_set)

    # Check if only user access related fields changed
    # Create copies without the user access fields for comparison
    current_without_access = {
        k: v
        for k, v in current_config.items()
        if k not in ['private', 'allowed_users']
    }
    new_without_access = {
        k: v
        for k, v in new_config.items()
        if k not in ['private', 'allowed_users']
    }

    only_user_access_changes = current_without_access == new_without_access

    return WorkspaceConfigComparison(
        only_user_access_changes=only_user_access_changes,
        private_changed=private_changed,
        private_old=private_old,
        private_new=private_new,
        allowed_users_changed=allowed_users_changed,
        allowed_users_old=allowed_users_old,
        allowed_users_new=allowed_users_new,
        removed_users=removed_users,
        added_users=added_users)


def _validate_workspace_config_changes_with_lock(
        workspace_name: str, current_config: Dict[str, Any],
        new_config: Dict[str, Any]) -> None:
    lock_id = backend_utils.workspace_lock_id(workspace_name)
    lock_timeout = backend_utils.WORKSPACE_LOCK_TIMEOUT_SECONDS
    try:
        with locks.get_lock(lock_id, lock_timeout):
            # Validate the configuration changes based on active resources
            _validate_workspace_config_changes(workspace_name, current_config,
                                               new_config)
    except locks.LockTimeout as e:
        raise RuntimeError(
            f'Failed to validate workspace {workspace_name!r} due to '
            'a timeout when trying to access database. Please '
            f'try again or manually remove the lock at {lock_id}. '
            f'{common_utils.format_exception(e)}') from None


def _validate_workspace_config_changes(workspace_name: str,
                                       current_config: Dict[str, Any],
                                       new_config: Dict[str, Any]) -> None:
    """Validate workspace configuration changes based on active resources.

    This function implements the logic:
    - If only allowed_users or private changed:
      - If private changed from true to false: allow it
      - If private changed from false to true: check that all active resources
        belong to allowed_users
      - If private didn't change: check that removed users don't have active
        resources
    - Otherwise: check that workspace has no active resources

    Args:
        workspace_name: The name of the workspace.
        current_config: The current workspace configuration.
        new_config: The new workspace configuration.

    Raises:
        ValueError: If the configuration change is not allowed due to active
        resources.
    """
    config_comparison = _compare_workspace_configs(current_config, new_config)

    if config_comparison.only_user_access_changes:
        # Only user access settings changed
        if config_comparison.private_changed:
            if (config_comparison.private_old and
                    not config_comparison.private_new):
                # Changed from private to public - always allow
                logger.info(
                    f'Workspace {workspace_name!r} changed from private to'
                    f' public.')
                return
            elif (not config_comparison.private_old and
                  config_comparison.private_new):
                # Changed from public to private - check that all active
                # resources belong to the new allowed users
                logger.info(
                    f'Workspace {workspace_name!r} changed from public to'
                    f' private. Checking that all active resources belong'
                    f' to allowed users.')

                error_summary, missed_users_names, _ = (
                    resource_checker.check_users_workspaces_active_resources(
                        config_comparison.allowed_users_new, [workspace_name]))
                if error_summary:
                    error_msg=f'Cannot change workspace {workspace_name!r}' \
                    f' to private '
                    if missed_users_names:
                        missed_users_list = ', '.join(missed_users_names)
                        if len(missed_users_names) == 1:
                            error_msg += f'because the user ' \
                            f'{missed_users_list!r} has {error_summary}'
                        else:
                            error_msg += f'because the users ' \
                            f'{missed_users_list!r} have {error_summary}'
                        error_msg += ' but not in the allowed_users list.' \
                        ' Please either add the users to allowed_users or' \
                        ' ask them to terminate their resources.'
                    raise ValueError(error_msg)
        else:
            # Private setting didn't change, but allowed_users changed
            if (config_comparison.allowed_users_changed and
                    config_comparison.removed_users):
                # Check that removed users don't have active resources
                logger.info(
                    f'Checking that removed users'
                    f' {config_comparison.removed_users} do not have'
                    f' active resources in workspace {workspace_name!r}.')
                error_summary, missed_users_names, missed_user_dict = (
                    resource_checker.check_users_workspaces_active_resources(
                        config_comparison.allowed_users_new, [workspace_name]))
                if error_summary:
                    error_user_ids = []
                    for user_id in config_comparison.removed_users:
                        if user_id in missed_user_dict:
                            error_user_ids.append(user_id)
                    error_user_names = []
                    if error_user_ids:
                        error_user_names = [
                            missed_user_dict[user_id]
                            for user_id in error_user_ids
                        ]

                    error_msg = 'Cannot '
                    error_users_list = ', '.join(error_user_names)
                    if len(error_user_names) == 1:
                        error_msg += f'remove user {error_users_list!r} ' \
                        f'from workspace {workspace_name!r} because the ' \
                        f'user has {error_summary}'
                    else:
                        error_msg += f'remove users {error_users_list!r}' \
                        f' from workspace {workspace_name!r} because the' \
                        f' users have {error_summary}'
                    error_msg += ', but not in the allowed_users list.' \
                    ' Please either add the users to allowed_users or' \
                    ' ask them to terminate their resources.'
                    raise ValueError(error_msg)
    else:
        # Other configuration changes - check that workspace has no active
        # resources
        logger.info(
            f'Non-user-access configuration changes detected for'
            f' workspace {workspace_name!r}. Checking that workspace has'
            f' no active resources.')
        resource_checker.check_no_active_resources_for_workspaces([
            (workspace_name, 'update')
        ])


@usage_lib.entrypoint
def update_workspace(workspace_name: str, config: Dict[str,
                                                       Any]) -> Dict[str, Any]:
    """Updates a specific workspace configuration.

    Args:
        workspace_name: The name of the workspace to update.
        config: The new configuration for the workspace.

    Returns:
        The updated workspaces configuration.

    Raises:
        ValueError: If the workspace configuration is invalid, or if there are
            active clusters or managed jobs that prevent the configuration
            change.
            The validation logic depends on what changed:
            - If only allowed_users or private changed:
              - Private true->false: Always allowed
              - Private false->true: All active resources must belong to
                allowed_users
              - allowed_users changes: Removed users must not have active
                resources
            - Other changes: Workspace must have no active resources
        FileNotFoundError: If the config file cannot be found.
        PermissionError: If the config file cannot be written.
    """
    _validate_workspace_config(workspace_name, config)

    # Get the current workspace configuration for comparison
    current_workspaces = skypilot_config.get_nested(('workspaces',),
                                                    default_value={})
    current_config = current_workspaces.get(workspace_name, {})

    _validate_workspace_config_changes_with_lock(workspace_name, current_config,
                                                 config)

    def update_workspace_fn(workspaces: Dict[str, Any]) -> None:
        """Function to update workspace inside the lock."""
        workspaces[workspace_name] = config
        users = workspaces_utils.get_workspace_users(config)
        permission_service = permission.permission_service
        permission_service.update_workspace_policy(workspace_name, users)

    # Use the internal helper function to save
    result = _update_workspaces_config(update_workspace_fn)

    # Validate the workspace by running sky check for it
    try:
        sky_check.check(quiet=True, workspace=workspace_name)
    except Exception as e:  # pylint: disable=broad-except
        logger.warning(f'Workspace {workspace_name} configuration saved but '
                       f'validation check failed: {e}')
        # Don't fail the update if the check fails, just warn

    return result


@usage_lib.entrypoint
def create_workspace(workspace_name: str, config: Dict[str,
                                                       Any]) -> Dict[str, Any]:
    """Creates a new workspace configuration.

    Args:
        workspace_name: The name of the workspace to create.
        config: The configuration for the new workspace.

    Returns:
        The updated workspaces configuration.

    Raises:
        ValueError: If the workspace already exists or configuration is invalid.
        FileNotFoundError: If the config file cannot be found.
        PermissionError: If the config file cannot be written.
    """
    # Validate the workspace name
    if not workspace_name or not isinstance(workspace_name, str):
        raise ValueError('Workspace name must be a non-empty string.')

    _validate_workspace_config(workspace_name, config)

    def create_workspace_fn(workspaces: Dict[str, Any]) -> None:
        """Function to create workspace inside the lock."""
        if workspace_name in workspaces:
            raise ValueError(f'Workspace {workspace_name!r} already exists. '
                             'Use update instead.')
        workspaces[workspace_name] = config
        # Add policy for the workspace and allowed users
        users = workspaces_utils.get_workspace_users(config)
        permission_service = permission.permission_service
        permission_service.add_workspace_policy(workspace_name, users)

    # Use the internal helper function to save
    result = _update_workspaces_config(create_workspace_fn)

    # Validate the workspace by running sky check for it
    try:
        sky_check.check(quiet=True, workspace=workspace_name)
    except Exception as e:  # pylint: disable=broad-except
        logger.warning(f'Workspace {workspace_name} configuration saved but '
                       f'validation check failed: {e}')
        # Don't fail the update if the check fails, just warn

    return result


@usage_lib.entrypoint
def delete_workspace(workspace_name: str) -> Dict[str, Any]:
    """Deletes a workspace configuration.

    Args:
        workspace_name: The name of the workspace to delete.

    Returns:
        The updated workspaces configuration.

    Raises:
        ValueError: If the workspace doesn't exist, is the default workspace,
                   or has active clusters or managed jobs.
        FileNotFoundError: If the config file cannot be found.
        PermissionError: If the config file cannot be written.
    """
    # Prevent deletion of default workspace
    if workspace_name == constants.SKYPILOT_DEFAULT_WORKSPACE:
        raise ValueError(f'Cannot delete the default workspace '
                         f'{constants.SKYPILOT_DEFAULT_WORKSPACE!r}.')

    # Check if workspace exists
    current_workspaces = get_workspaces()
    if workspace_name not in current_workspaces:
        raise ValueError(f'Workspace {workspace_name!r} does not exist.')

    # Check for active clusters and managed jobs in the workspace
    resource_checker.check_no_active_resources_for_workspaces([(workspace_name,
                                                                'delete')])

    def delete_workspace_fn(workspaces: Dict[str, Any]) -> None:
        """Function to delete workspace inside the lock."""
        if workspace_name not in workspaces:
            raise ValueError(f'Workspace {workspace_name!r} does not exist.')
        del workspaces[workspace_name]
        permission_service = permission.permission_service
        permission_service.remove_workspace_policy(workspace_name)

    # Use the internal helper function to save
    return _update_workspaces_config(delete_workspace_fn)


# =========================
# = Config Management =
# =========================


@usage_lib.entrypoint
def get_config() -> Dict[str, Any]:
    """Returns the entire SkyPilot configuration.

    Returns:
        The complete SkyPilot configuration as a dictionary.
    """
    return skypilot_config.to_dict()


@usage_lib.entrypoint
def update_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """Updates the entire SkyPilot configuration.

    Args:
        config: The new configuration to save.

    Returns:
        The updated configuration.

    Raises:
        ValueError: If the configuration is invalid, or if there are
            active clusters or managed jobs in workspaces being modified.
        FileNotFoundError: If the config file cannot be found.
        PermissionError: If the config file cannot be written.
    """
    # Validate the configuration using the schema
    try:
        common_utils.validate_schema(config, schemas.get_config_schema(),
                                     'Invalid SkyPilot configuration: ')
    except exceptions.InvalidSkyPilotConfigError as e:
        raise ValueError(str(e)) from e

    # Check for API server changes and validate them
    current_config = skypilot_config.to_dict()
    # If there is no changes to the config, we can return early
    if current_config == config:
        return config

    current_endpoint = current_config.get('api_server', {}).get('endpoint')
    new_endpoint = config.get('api_server', {}).get('endpoint')
    if current_endpoint != new_endpoint:
        raise ValueError('API server endpoint should not be changed to avoid '
                         'unexpected behavior.')

    # Check for workspace changes and validate them
    current_workspaces = current_config.get('workspaces', {})
    new_workspaces = config.get('workspaces', {})

    # Collect all workspaces that need to be checked for active resources
    workspaces_to_check: List[Tuple[str, str]] = []
    workspaces_to_check_policy: Dict[str, Dict[str, List[str]]] = {
        'add': {},
        'update': {},
        'delete': {}
    }

    # Check each workspace that is being modified
    for workspace_name, new_workspace_config in new_workspaces.items():
        if workspace_name not in current_workspaces:
            users = workspaces_utils.get_workspace_users(new_workspace_config)
            workspaces_to_check_policy['add'][workspace_name] = users
            continue

        current_workspace_config = current_workspaces.get(workspace_name, {})

        # If workspace configuration is changing, validate and mark for checking
        if current_workspace_config != new_workspace_config:
            _validate_workspace_config(workspace_name, new_workspace_config)
            _validate_workspace_config_changes_with_lock(
                workspace_name, current_workspace_config, new_workspace_config)
            users = workspaces_utils.get_workspace_users(new_workspace_config)
            workspaces_to_check_policy['update'][workspace_name] = users

    # Check for workspace deletions
    for workspace_name in current_workspaces:
        if workspace_name not in new_workspaces:
            # Workspace is being deleted
            if workspace_name == constants.SKYPILOT_DEFAULT_WORKSPACE:
                raise ValueError(f'Cannot delete the default workspace '
                                 f'{constants.SKYPILOT_DEFAULT_WORKSPACE!r}.')
            workspaces_to_check.append((workspace_name, 'delete'))
            workspaces_to_check_policy['delete'][workspace_name] = ['*']

    # Check all workspaces for active resources in one efficient call
    resource_checker.check_no_active_resources_for_workspaces(
        workspaces_to_check)

    # Use file locking to prevent race conditions
    lock_path = skypilot_config.get_skypilot_config_lock_path()
    try:
        with filelock.FileLock(lock_path,
                               _WORKSPACE_CONFIG_LOCK_TIMEOUT_SECONDS):
            # Convert to config_utils.Config and save
            config_obj = config_utils.Config.from_dict(config)
            skypilot_config.update_api_server_config_no_lock(config_obj)
            permission_service = permission.permission_service
            for operation, workspaces in workspaces_to_check_policy.items():
                for workspace_name, users in workspaces.items():
                    if operation == 'add':
                        permission_service.add_workspace_policy(
                            workspace_name, users)
                    elif operation == 'update':
                        permission_service.update_workspace_policy(
                            workspace_name, users)
                    elif operation == 'delete':
                        permission_service.remove_workspace_policy(
                            workspace_name)
    except filelock.Timeout as e:
        raise RuntimeError(
            f'Failed to update configuration due to a timeout '
            f'when trying to acquire the lock at {lock_path}. This may '
            'indicate another SkyPilot process is currently updating the '
            'configuration. Please try again or manually remove the lock '
            f'file if you believe it is stale.') from e

    # Validate the configuration by running sky check
    try:
        sky_check.check(quiet=True)
    except Exception as e:  # pylint: disable=broad-except
        logger.warning(f'Configuration saved but '
                       f'validation check failed: {e}')
        # Don't fail the update if the check fails, just warn

    return config


def reject_request_for_unauthorized_workspace(user: models.User) -> None:
    """Rejects a request that has no permission to access active workspace.

    Args:
        user: The user making the request.

    Raises:
        PermissionDeniedError: If the user does not have permission to access
            the active workspace.
    """
    active_workspace = skypilot_config.get_active_workspace()
    if not permission.permission_service.check_workspace_permission(
            user.id, active_workspace):
        raise exceptions.PermissionDeniedError(
            f'User {user.name} ({user.id}) does not have '
            f'permission to access workspace {active_workspace!r}')


def is_workspace_private(workspace_config: Dict[str, Any]) -> bool:
    """Check if a workspace is private.

    Args:
        workspace_config: The workspace configuration dictionary.

    Returns:
        True if the workspace is private, False if it's public.
    """
    return workspace_config.get('private', False)


@annotations.lru_cache(scope='request', maxsize=1)
def workspaces_for_user(user_id: str) -> Dict[str, Any]:
    """Returns the workspaces that the user has access to.

    Args:
        user_id: The user id to check.

    Returns:
        A map from workspace name to workspace configuration.
    """
    workspaces = skypilot_config.get_nested(('workspaces',), default_value={})
    if constants.SKYPILOT_DEFAULT_WORKSPACE not in workspaces:
        workspaces[constants.SKYPILOT_DEFAULT_WORKSPACE] = {}
    user_workspaces = {}

    for workspace_name, workspace_config in workspaces.items():
        if permission.permission_service.check_workspace_permission(
                user_id, workspace_name):
            user_workspaces[workspace_name] = workspace_config
    return user_workspaces
